"""
Init command to scaffold a project app from a template
"""
import json
import logging
from json import JSONDecodeError

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import common_options, pass_context, print_cmdline_args
from samcli.commands._utils.click_mutex import ClickMutex
from samcli.commands.init.init_flow_helpers import _get_runtime_from_image, get_architectures, get_sorted_runtimes
from samcli.lib.build.constants import DEPRECATED_RUNTIMES
from samcli.lib.telemetry.metric import track_command
from samcli.lib.utils.architecture import ARM64, X86_64
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.lib.utils.version_checker import check_newer_version
from samcli.local.common.runtime_template import INIT_RUNTIMES, LAMBDA_IMAGES_RUNTIMES, SUPPORTED_DEP_MANAGERS

LOG = logging.getLogger(__name__)

HELP_TEXT = """ \b
    Initialize a serverless application with a SAM template, folder
    structure for your Lambda functions, connected to an event source such as APIs,
    S3 Buckets or DynamoDB Tables. This application includes everything you need to
    get started with serverless and eventually grow into a production scale application.
    \b
    This command can initialize a boilerplate serverless app. If you want to create your own
    template as well as use a custom location please take a look at our official documentation.
\b
Common usage:
    \b
    Starts an interactive prompt process to initialize a new project:
    \b
    $ sam init
    \b
    Initializes a new SAM project using project templates without an interactive workflow:
    \b
    $ sam init --name sam-app --runtime nodejs14.x --dependency-manager npm --app-template hello-world
    \b
    $ sam init --name sam-app --runtime nodejs14.x --architecture arm64
    \b
    $ sam init --name sam-app --package-type image --base-image nodejs14.x-base
    \b
    Initializes a new SAM project using custom template in a Git/Mercurial repository
    \b
    # gh being expanded to github url
    $ sam init --location gh:aws-samples/cookiecutter-aws-sam-python
    \b
    $ sam init --location git+ssh://git@github.com/aws-samples/cookiecutter-aws-sam-python.git
    \b
    $ sam init --location hg+ssh://hg@bitbucket.org/repo/template-name
    \b
    Initializes a new SAM project using custom template in a Zipfile
    \b
    $ sam init --location /path/to/template.zip
    \b
    $ sam init --location https://example.com/path/to/template.zip
    \b
    Initializes a new SAM project using custom template in a local path
    \b
    $ sam init --location /path/to/template/folder
"""

INCOMPATIBLE_PARAMS_HINT = """You can run 'sam init' without any options for an interactive initialization flow, \
or you can provide one of the following required parameter combinations:
\t--name, --location, or
\t--name, --package-type, --base-image, or
\t--name, --runtime, --app-template, --dependency-manager
"""

REQUIRED_PARAMS_HINT = "You can also re-run without the --no-interactive flag to be prompted for required values."

INIT_INTERACTIVE_OPTION_GUIDE = """
You can preselect a particular runtime or package type when using the `sam init` experience.
Call `sam init --help` to learn more.
"""


class PackageType:
    """
    This class has a callback function for the --package-type parameter to handle default value
    and also store if the --package-type param was passed explicitly
    """

    explicit = False

    def __init__(self):
        pass

    @staticmethod
    def pt_callback(ctx, param, provided_value):
        """
        This function is the callback for the --package-type param. Here we check if --package-type was passed or not.
        If not, we use the default value of --package-type to be Zip.
        """
        if provided_value is None:
            return ZIP
        PackageType.explicit = True
        return provided_value


def non_interactive_validation(func):
    """
    Check requirement for --dependency-manager parameter for non interactive mode

    --dependency-manager parameter is only required if --package-type is ZIP
    or --base-image is one of the java ones
    """

    def wrapped(*args, **kwargs):
        ctx = click.get_current_context()
        non_interactive = ctx.params.get("no_interactive")

        # only run in non interactive mode
        if non_interactive:
            package_type = ctx.params.get("package_type")
            base_image = ctx.params.get("base_image")
            dependency_manager = ctx.params.get("dependency_manager")
            location = ctx.params.get("location")

            # TODO: improve how we check for java type images instead of searching for substring
            java_base_image = base_image and "java" in base_image

            # dependency manager is only required for ZIP types if location is not also specified
            # and is required for java IMAGE packages
            if not location and (package_type == ZIP or java_base_image):
                if not dependency_manager:
                    raise click.UsageError("Missing parameter --dependency-manager")

        return func(*args, **kwargs)

    return wrapped


@click.command(
    "init",
    help=HELP_TEXT,
    short_help="Init an AWS SAM application.",
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@configuration_option(provider=TomlProvider(section="parameters"))
@click.option(
    "--no-interactive",
    is_flag=True,
    default=False,
    help="Disable interactive prompting for init parameters, and fail if any required values are missing.",
    cls=ClickMutex,
    required_param_lists=[
        ["name", "location"],
        ["name", "package_type", "base_image"],
        ["name", "runtime", "dependency_manager", "app_template"],
        # check non_interactive_validation for additional validations
    ],
    required_params_hint=REQUIRED_PARAMS_HINT,
)
@click.option(
    "-a",
    "--architecture",
    type=click.Choice([ARM64, X86_64]),
    help="Architectures your Lambda function will run on",
    cls=ClickMutex,
)
@click.option(
    "-l",
    "--location",
    help="Template location (git, mercurial, http(s), zip, path)",
    cls=ClickMutex,
    incompatible_params=["package_type", "runtime", "base_image", "dependency_manager", "app_template"],
    incompatible_params_hint=INCOMPATIBLE_PARAMS_HINT,
)
@click.option(
    "-r",
    "--runtime",
    type=click.Choice(get_sorted_runtimes(INIT_RUNTIMES)),
    help="Lambda Runtime of your app",
    cls=ClickMutex,
    incompatible_params=["location", "base_image"],
    incompatible_params_hint=INCOMPATIBLE_PARAMS_HINT,
)
@click.option(
    "-p",
    "--package-type",
    type=click.Choice([ZIP, IMAGE]),
    help="Package type for your app",
    cls=ClickMutex,
    callback=PackageType.pt_callback,
    incompatible_params=["location"],
    incompatible_params_hint=INCOMPATIBLE_PARAMS_HINT,
)
@click.option(
    "-i",
    "--base-image",
    type=click.Choice(LAMBDA_IMAGES_RUNTIMES),
    default=None,
    help="Lambda Image of your app",
    cls=ClickMutex,
    incompatible_params=["location", "runtime"],
    incompatible_params_hint=INCOMPATIBLE_PARAMS_HINT,
)
@click.option(
    "-d",
    "--dependency-manager",
    type=click.Choice(SUPPORTED_DEP_MANAGERS),
    default=None,
    help="Dependency manager of your Lambda runtime",
    required=False,
    cls=ClickMutex,
    incompatible_params=["location"],
    incompatible_params_hint=INCOMPATIBLE_PARAMS_HINT,
)
@click.option("-o", "--output-dir", type=click.Path(), help="Where to output the initialized app into", default=".")
@click.option("-n", "--name", help="Name of your project to be generated as a folder")
@click.option(
    "--app-template",
    help="Identifier of the managed application template you want to use. "
    "If not sure, call 'sam init' without options for an interactive workflow.",
    cls=ClickMutex,
    incompatible_params=["location"],
    incompatible_params_hint=INCOMPATIBLE_PARAMS_HINT,
)
@click.option(
    "--no-input",
    is_flag=True,
    default=False,
    help="Disable Cookiecutter prompting and accept default values defined template config",
)
@click.option(
    "--extra-context",
    default=None,
    help="Override any custom parameters in the template's cookiecutter.json configuration e.g. "
    ""
    '{"customParam1": "customValue1", "customParam2":"customValue2"}'
    """ """,
    required=False,
)
@click.option(
    "--tracing/--no-tracing",
    default=None,
    help="Enable AWS X-Ray tracing for your lambda functions",
)
@click.option(
    "--application-insights/--no-application-insights",
    default=None,
    help="Enable CloudWatch Application Insights monitoring for your application",
)
@common_options
@non_interactive_validation
@pass_context
@track_command
@check_newer_version
@print_cmdline_args
def cli(
    ctx,
    no_interactive,
    location,
    package_type,
    runtime,
    architecture,
    base_image,
    dependency_manager,
    output_dir,
    name,
    app_template,
    no_input,
    extra_context,
    tracing,
    application_insights,
    config_file,
    config_env,
):
    """
    `sam init` command entry point
    """
    do_cli(
        ctx,
        no_interactive,
        location,
        PackageType.explicit,
        package_type,
        runtime,
        architecture,
        base_image,
        dependency_manager,
        output_dir,
        name,
        app_template,
        no_input,
        extra_context,
        tracing,
        application_insights,
    )  # pragma: no cover


# pylint: disable=too-many-locals
def do_cli(
    ctx,
    no_interactive,
    location,
    pt_explicit,
    package_type,
    runtime,
    architecture,
    base_image,
    dependency_manager,
    output_dir,
    name,
    app_template,
    no_input,
    extra_context,
    tracing,
    application_insights,
):
    """
    Implementation of the ``cli`` method
    """

    from samcli.commands.exceptions import LambdaImagesTemplateException
    from samcli.commands.init.init_generator import do_generate
    from samcli.commands.init.init_templates import InitTemplates
    from samcli.commands.init.interactive_init_flow import do_interactive

    _deprecate_notification(runtime)

    # check for required parameters
    zip_bool = name and runtime and dependency_manager and app_template
    image_bool = name and pt_explicit and base_image
    if location or zip_bool or image_bool:
        # need to turn app_template into a location before we generate
        templates = InitTemplates()
        if package_type == IMAGE and image_bool:
            runtime = _get_runtime_from_image(base_image)
            options = templates.init_options(package_type, runtime, base_image, dependency_manager)
            if not app_template:
                if len(options) == 1:
                    app_template = options[0].get("appTemplate")
                elif len(options) > 1:
                    raise LambdaImagesTemplateException(
                        "Multiple lambda image application templates found. "
                        "Please specify one using the --app-template parameter."
                    )

        if app_template and not location:
            location = templates.location_from_app_template(
                package_type, runtime, base_image, dependency_manager, app_template
            )
            no_input = True
        extra_context = _get_cookiecutter_template_context(name, runtime, architecture, extra_context)

        if not output_dir:
            output_dir = "."
        do_generate(
            location,
            package_type,
            runtime,
            dependency_manager,
            output_dir,
            name,
            no_input,
            extra_context,
            tracing,
            application_insights,
        )
    else:
        if not (pt_explicit or runtime or dependency_manager or base_image or architecture):
            click.secho(INIT_INTERACTIVE_OPTION_GUIDE, fg="yellow", bold=True)

        # proceed to interactive state machine, which will call do_generate
        do_interactive(
            location,
            pt_explicit,
            package_type,
            runtime,
            architecture,
            base_image,
            dependency_manager,
            output_dir,
            name,
            app_template,
            no_input,
            tracing,
            application_insights,
        )


def _deprecate_notification(runtime):
    from samcli.lib.utils.colors import Colored

    if runtime in DEPRECATED_RUNTIMES:
        message = (
            f"WARNING: {runtime} is no longer supported by AWS Lambda, please update to a newer supported runtime. "
            "For more information please check AWS Lambda Runtime Support Policy: "
            "https://docs.aws.amazon.com/lambda/latest/dg/runtime-support-policy.html"
        )
        LOG.warning(Colored().yellow(message))


def _get_cookiecutter_template_context(name, runtime, architecture, extra_context):
    default_context = {}
    extra_context_dict = {}

    if runtime is not None:
        default_context["runtime"] = runtime

    if name is not None:
        default_context["project_name"] = name

    default_context["architectures"] = {"value": get_architectures(architecture)}
    if extra_context is not None:
        try:
            extra_context_dict = json.loads(extra_context)
        except JSONDecodeError as ex:
            raise click.UsageError(
                "Parse error reading the --extra-context parameter. The value of this parameter must be valid JSON."
            ) from ex

    return {**extra_context_dict, **default_context}



""" Contains the data types used in the TF prepare hook"""
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Union


@dataclass
class ConstantValue:
    value: Any


@dataclass
class References:
    value: List[str]


Expression = Union[ConstantValue, References]


@dataclass
class ResolvedReference:
    value: str
    # root module address is None
    module_address: Optional[str]


@dataclass
class TFModule:
    # full path to the module, including parent modules
    full_address: Optional[str]
    parent_module: Optional["TFModule"]
    variables: Dict[str, Expression]
    resources: Dict[str, "TFResource"]
    child_modules: Dict[str, "TFModule"]
    outputs: Dict[str, Expression]

    # current module's + all child modules' resources
    def get_all_resources(self) -> List["TFResource"]:
        all_resources = [deepcopy(resource) for resource in self.resources.values()]
        for _, module in self.child_modules.items():
            all_resources += module.get_all_resources()

        return all_resources

    @property
    def module_name(self):
        return self.full_address if self.full_address else "root module"


@dataclass
class TFResource:
    address: str
    type: str
    # the module this resource is defined in
    module: TFModule
    attributes: Dict[str, Expression]

    @property
    def full_address(self) -> str:
        if self.module and self.module.full_address:
            return f"{self.module.full_address}.{self.address}"
        return self.address


PropertyBuilder = Callable[[dict, TFResource], Any]
PropertyBuilderMapping = Dict[str, PropertyBuilder]


@dataclass
class ResourceTranslator:
    cfn_name: str
    property_builder_mapping: PropertyBuilderMapping


@dataclass
class SamMetadataResource:
    current_module_address: Optional[str]
    resource: Dict
    config_resource: TFResource



"""
The producer for the 'sam list stack-outputs' command
"""
import dataclasses
import logging
from typing import Any, Optional

from botocore.exceptions import BotoCoreError, ClientError

from samcli.commands.list.exceptions import (
    NoOutputsForStackError,
    SamListUnknownBotoCoreError,
    SamListUnknownClientError,
    StackDoesNotExistInRegionError,
)
from samcli.lib.list.list_interfaces import Producer
from samcli.lib.list.stack_outputs.stack_outputs import StackOutputs
from samcli.lib.utils.boto_utils import get_client_error_code

LOG = logging.getLogger(__name__)


class StackOutputsProducer(Producer):
    def __init__(self, stack_name, output, region, cloudformation_client, mapper, consumer):
        self.stack_name = stack_name
        self.output = output
        self.region = region
        self.cloudformation_client = cloudformation_client
        self.mapper = mapper
        self.consumer = consumer

    def get_stack_info(self) -> Optional[Any]:
        """
        Returns the stack output information for the stack and raises exceptions accordingly

        Returns
        -------
            A dictionary containing the stack's information
        """

        try:
            response = self.cloudformation_client.describe_stacks(StackName=self.stack_name)
            if not response.get("Stacks", []):
                raise StackDoesNotExistInRegionError(stack_name=self.stack_name, region=self.region)
            if len(response.get("Stacks", [])) > 0 and "Outputs" not in response.get("Stacks", [])[0]:
                raise NoOutputsForStackError(stack_name=self.stack_name, region=self.region)
            return response["Stacks"][0]["Outputs"]

        except ClientError as e:
            if get_client_error_code(e) == "ValidationError":
                LOG.debug("Stack with id %s does not exist", self.stack_name)
                raise StackDoesNotExistInRegionError(stack_name=self.stack_name, region=self.region) from e
            LOG.error("ClientError Exception : %s", str(e))
            raise SamListUnknownClientError(msg=str(e)) from e
        except BotoCoreError as e:
            LOG.error("Botocore Exception : %s", str(e))
            raise SamListUnknownBotoCoreError(msg=str(e)) from e

    def produce(self):
        response = self.get_stack_info()
        output_list = []
        for stack_output in response:
            stack_output_data = StackOutputs(
                OutputKey=stack_output.get("OutputKey", ""),
                OutputValue=stack_output.get("OutputValue", ""),
                Description=stack_output.get("Description", ""),
            )
            output_list.append(dataclasses.asdict(stack_output_data))
        mapped_output = self.mapper.map(output_list)
        self.consumer.consume(data=mapped_output)



"""
Plugins are sub-sets of templates, it encapsulate common behavior of different templates and plugged to each of them
"""
from typing import NamedTuple

from .interactive_flow import InteractiveFlow
from .processor import Processor


class Plugin(NamedTuple):
    """
    A plugin is a sub cookiecutter template. it has its own interactive_flow to prompt the user for answers to
    its own context and it also has its preprocessor and postprocessor.  plugin's components are appended to the
    templates corresponding component; interactive_flows, preprocessors and postprocessors.
    plugins encapsulate common logic of different templates in one place to be passed(plugged) to each template.

    Attributes
    ----------
    interactive_flow: InteractiveFlow
        a flow of questions to be appended to the series of interactive_flows of the parent template.
    preprocessor: Processor
        a processor to be appended to the series of processors of the parent template.
    postprocessor: Processor
        a processor to be appended to the series of postprocessors of the parent template.
    """

    interactive_flow: InteractiveFlow
    preprocessor: Processor
    postprocessor: Processor



"""Class that parses the CloudFormation Api Template"""
import logging
from typing import Any, Dict, List, Optional, Type, Union

from samcli.commands.local.lib.swagger.parser import SwaggerParser
from samcli.commands.local.lib.swagger.reader import SwaggerReader
from samcli.commands.validate.lib.exceptions import InvalidSamDocumentException
from samcli.lib.providers.api_collector import ApiCollector
from samcli.lib.providers.provider import (
    CORS_CREDENTIALS_HEADER,
    CORS_HEADERS_HEADER,
    CORS_MAX_AGE_HEADER,
    CORS_METHODS_HEADER,
    CORS_ORIGIN_HEADER,
    Cors,
    Stack,
)
from samcli.local.apigw.local_apigw_service import Route

LOG = logging.getLogger(__name__)

ALLOW_HEADERS = "AllowHeaders"
ALLOW_ORIGIN = "AllowOrigin"
ALLOW_METHODS = "AllowMethods"
ALLOW_CREDENTIALS = "AllowCredentials"
MAX_AGE = "MaxAge"


class CfnBaseApiProvider:
    RESOURCE_TYPE = "Type"

    def extract_resources(self, stacks: List[Stack], collector: ApiCollector, cwd: Optional[str] = None) -> None:
        """
        Extract the Route Object from a given resource and adds it to the RouteCollector.

        Parameters
        ----------
        stacks: List[Stack]
            List of stacks apis are extracted from
        collector: samcli.lib.providers.api_collector.ApiCollector
            Instance of the API collector that where we will save the API information
        cwd : str
            Optional working directory with respect to which we will resolve relative path to Swagger file
        """
        raise NotImplementedError("not implemented")

    @staticmethod
    def extract_swagger_route(
        stack_path: str,
        logical_id: str,
        body: Dict,
        uri: Union[str, Dict],
        binary_media: Optional[List],
        collector: ApiCollector,
        cwd: Optional[str] = None,
        event_type: str = Route.API,
    ) -> None:
        """
        Parse the Swagger documents and adds it to the ApiCollector.

        Parameters
        ----------
        stack_path : str
            Path of the stack the resource is located

        logical_id : str
            Logical ID of the resource
        body : dict
            The body of the RestApi
        uri : str or dict
            The url to location of the RestApi
        binary_media : list
            The link to the binary media
        collector : samcli.lib.providers.api_collector.ApiCollector
            Instance of the Route collector that where we will save the route information
        cwd : str
            Optional working directory with respect to which we will resolve relative path to Swagger file
        event_type : str
            The event type, 'Api' or 'HttpApi', see samcli/local/apigw/local_apigw_service.py:35
        """
        reader = SwaggerReader(definition_body=body, definition_uri=uri, working_dir=cwd)
        swagger = reader.read()
        parser = SwaggerParser(stack_path, swagger)
        routes = parser.get_routes(event_type)
        LOG.debug("Found '%s' APIs in resource '%s'", len(routes), logical_id)

        collector.add_routes(logical_id, routes)

        collector.add_binary_media_types(logical_id, parser.get_binary_media_types())  # Binary media from swagger
        collector.add_binary_media_types(logical_id, binary_media)  # Binary media specified on resource in template

    def extract_cors(self, cors_prop: Union[Dict, str]) -> Optional[Cors]:
        """
        Extract Cors property from AWS::Serverless::Api resource by reading and parsing Swagger documents. The result
        is added to the Api.

        Parameters
        ----------
        cors_prop : dict
            Resource properties for Cors
        """
        cors = None
        if cors_prop and isinstance(cors_prop, dict):
            allow_methods = self._get_cors_prop(cors_prop, ALLOW_METHODS)
            if allow_methods:
                allow_methods = CfnBaseApiProvider.normalize_cors_allow_methods(allow_methods)
            else:
                allow_methods = ",".join(sorted(Route.ANY_HTTP_METHODS))

            allow_origin = self._get_cors_prop(cors_prop, ALLOW_ORIGIN)
            allow_headers = self._get_cors_prop(cors_prop, ALLOW_HEADERS)
            allow_credentials = self._get_cors_prop(cors_prop, ALLOW_CREDENTIALS, True)
            max_age = self._get_cors_prop(cors_prop, MAX_AGE)

            cors = Cors(
                allow_origin=allow_origin,
                allow_methods=allow_methods,
                allow_headers=allow_headers,
                allow_credentials=allow_credentials,
                max_age=max_age,
            )
        elif cors_prop and isinstance(cors_prop, str):
            allow_origin = cors_prop
            if not (allow_origin.startswith("'") and allow_origin.endswith("'")):
                raise InvalidSamDocumentException(
                    "Cors Properties must be a quoted string " '(i.e. "\'*\'" is correct, but "*" is not).'
                )
            allow_origin = allow_origin.strip("'")

            cors = Cors(
                allow_origin=allow_origin,
                allow_methods=",".join(sorted(Route.ANY_HTTP_METHODS)),
                allow_headers=None,
                allow_credentials=None,
                max_age=None,
            )
        return cors

    def extract_cors_from_method(self, cors_dict: Dict) -> Optional[Cors]:
        """
        Extract the cors parameters from an AWS::ApiGateway::Method.
        Parameters
        ----------
        cors_dict : dict
            A dict of cors parameters in the format of ResponseParameters
        Return
        ------
        A Cors object with the relevant cors headers from the ResponseParameters
        """
        cors_props = {
            ALLOW_HEADERS: cors_dict.get(CfnBaseApiProvider._get_response_header(CORS_HEADERS_HEADER)),
            ALLOW_METHODS: cors_dict.get(CfnBaseApiProvider._get_response_header(CORS_METHODS_HEADER)),
            ALLOW_ORIGIN: cors_dict.get(CfnBaseApiProvider._get_response_header(CORS_ORIGIN_HEADER)),
            ALLOW_CREDENTIALS: cors_dict.get(CfnBaseApiProvider._get_response_header(CORS_CREDENTIALS_HEADER)),
            MAX_AGE: cors_dict.get(CfnBaseApiProvider._get_response_header(CORS_MAX_AGE_HEADER)),
        }
        if all(value is None for value in cors_props.values()):
            return None
        return self.extract_cors(cors_props)

    @staticmethod
    def _get_response_header(allow_method: str) -> str:
        """
        Get the full response header key with the cors method type.
        Parameters
        ----------
        allow_method : str
            The type of cors header
        Return
        ------
        A string matching the ResponseParameters key format
        """
        return "method.response.header." + allow_method

    @staticmethod
    def _get_cors_prop(cors_dict: Dict, prop_name: str, allow_bool: bool = False) -> Optional[str]:
        """
        Extract cors properties from dictionary and remove extra quotes.

        Parameters
        ----------
        cors_dict : dict
            Resource properties for Cors

        prop_name : str
            Cors property to get the value for

        allow_bool : bool
            If a boolean value is allowed for this property or not (defaults to false)

        Return
        ------
        A string with the extra quotes removed
        """
        prop = cors_dict.get(prop_name)
        if prop:
            if allow_bool and isinstance(prop, bool):
                prop = "'true'"  # We alredy know this is true due to L141 passing
            if not isinstance(prop, str) or prop.startswith("!"):
                LOG.warning(
                    "CORS Property %s was not fully resolved. Will proceed as if the Property was not defined.",
                    prop_name,
                )
                return None

            if not (prop.startswith("'") and prop.endswith("'")):
                raise InvalidSamDocumentException(
                    "{} must be a quoted string " '(i.e. "\'value\'" is correct, but "value" is not).'.format(prop_name)
                )
            prop = prop.strip("'")
        return prop

    def extract_cors_http(
        self,
        cors_prop: Union[bool, Dict],
    ) -> Optional[Cors]:
        """
        Extract Cors property from AWS::Serverless::HttpApi resource by reading and parsing Swagger documents.
        The result is added to the HttpApi.

        Parameters
        ----------
        cors_prop : dict
            Resource properties for CorsConfiguration
        """
        cors = None
        if cors_prop and isinstance(cors_prop, dict):
            allow_methods = self._get_cors_prop_http(cors_prop, "AllowMethods", list)
            if isinstance(allow_methods, list):
                allow_methods = CfnBaseApiProvider.normalize_cors_allow_methods(allow_methods)
            else:
                allow_methods = ",".join(sorted(Route.ANY_HTTP_METHODS))

            allow_origins = self._get_cors_prop_http(cors_prop, "AllowOrigins", list)
            if isinstance(allow_origins, list):
                allow_origins = ",".join(allow_origins)
            allow_headers = self._get_cors_prop_http(cors_prop, "AllowHeaders", list)
            if isinstance(allow_headers, list):
                allow_headers = ",".join(allow_headers)

            # Read AllowCredentials but only output the header with the case-sensitive value of true
            # (see https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Credentials)
            allow_credentials = "true" if self._get_cors_prop_http(cors_prop, "AllowCredentials", bool) else None

            max_age = self._get_cors_prop_http(cors_prop, "MaxAge", int)

            cors = Cors(
                allow_origin=allow_origins,
                allow_methods=allow_methods,
                allow_headers=allow_headers,
                allow_credentials=allow_credentials,
                max_age=max_age,
            )
        elif cors_prop and isinstance(cors_prop, bool) and cors_prop:
            cors = Cors(
                allow_origin="*",
                allow_methods=",".join(sorted(Route.ANY_HTTP_METHODS)),
                allow_headers=None,
                allow_credentials=None,
                max_age=None,
            )
        return cors

    @staticmethod
    def _get_cors_prop_http(
        cors_dict: Dict,
        prop_name: str,
        expect_type: Type,
    ) -> Optional[Any]:
        """
        Extract cors properties from dictionary.

        Parameters
        ----------
        cors_dict : dict
            Resource properties for Cors
        prop_name : str
            Property name
        expect_type : type
            Expect property type

        Return
        ------
        Value with matching type
        """
        prop = cors_dict.get(prop_name)
        if prop:
            if not isinstance(prop, expect_type):
                LOG.warning(
                    "CORS Property %s was not fully resolved. Will proceed as if the Property was not defined.",
                    prop_name,
                )
                return None
        return prop

    @staticmethod
    def normalize_cors_allow_methods(allow_methods: Union[str, List[str]]) -> str:
        """
        Normalize cors AllowMethods and Options to the methods if it's missing.

        Parameters
        ----------
        allow_methods : str
            The allow_methods string provided in the query

        Return
        -------
        A string with normalized route
        """
        if allow_methods == "*" or (isinstance(allow_methods, list) and "*" in allow_methods):
            return ",".join(sorted(Route.ANY_HTTP_METHODS))
        if isinstance(allow_methods, list):
            methods = allow_methods
        else:
            methods = allow_methods.split(",")
        normalized_methods = []
        for method in methods:
            normalized_method = method.strip().upper()
            if normalized_method not in Route.ANY_HTTP_METHODS:
                raise InvalidSamDocumentException("The method {} is not a valid CORS method".format(normalized_method))
            normalized_methods.append(normalized_method)

        if "OPTIONS" not in normalized_methods:
            normalized_methods.append("OPTIONS")

        return ",".join(sorted(normalized_methods))



"""
Contains all mappers (formatters) for CloudWatch logs
"""
import json
import logging
from json import JSONDecodeError
from typing import Any

from samcli.lib.observability.cw_logs.cw_log_event import CWLogEvent
from samcli.lib.observability.observability_info_puller import ObservabilityEventMapper
from samcli.lib.utils.colors import Colored
from samcli.lib.utils.time import timestamp_to_iso

LOG = logging.getLogger(__name__)


class CWKeywordHighlighterFormatter(ObservabilityEventMapper[CWLogEvent]):
    """
    Mapper implementation which will highlight given keywords in CloudWatch logs
    """

    def __init__(self, colored: Colored, keyword=None):
        """
        Parameters
        ----------
        colored : Colored
            Colored class that will be used to highlight the keywords in log event
        keyword : str
            Keyword that will be highlighted
        """
        self._keyword = keyword
        self._colored = colored

    def map(self, event: CWLogEvent) -> CWLogEvent:
        if self._keyword:
            highlight = self._colored.underline(self._keyword)
            event.message = event.message.replace(self._keyword, highlight)

        return event


class CWColorizeErrorsFormatter(ObservabilityEventMapper[CWLogEvent]):
    """
    Mapper implementation which will colorize some pre-defined error messages
    """

    # couple of pre-defined error messages for lambda functions which will be colorized when getting the logs
    NODEJS_CRASH_MESSAGE = "Process exited before completing request"
    TIMEOUT_MSG = "Task timed out"

    def __init__(self, colored: Colored):
        self._colored = colored

    def map(self, event: CWLogEvent) -> CWLogEvent:
        if (
            CWColorizeErrorsFormatter.NODEJS_CRASH_MESSAGE in event.message
            or CWColorizeErrorsFormatter.TIMEOUT_MSG in event.message
        ):
            event.message = self._colored.red(event.message)
        return event


class CWJsonFormatter(ObservabilityEventMapper[CWLogEvent]):
    """
    Mapper implementation which will auto indent the input if the input is a JSON object
    """

    # pylint: disable=R0201
    # Pylint recommends converting this method to a static one but we want it to stay as it is
    # since formatters/mappers are combined in an array of ObservabilityEventMapper class
    def map(self, event: CWLogEvent) -> CWLogEvent:
        try:
            if event.message.startswith("{"):
                msg_dict = json.loads(event.message)
                event.message = json.dumps(msg_dict, indent=2)
        except JSONDecodeError as err:
            LOG.debug("Can't decode string (%s) as JSON. Error (%s)", event.message, err)

        return event


class CWPrettyPrintFormatter(ObservabilityEventMapper[CWLogEvent]):
    """
    Mapper implementation which will format given CloudWatch log event into string with coloring
    log stream name and timestamp
    """

    def __init__(self, colored: Colored):
        self._colored = colored

    def map(self, event: CWLogEvent) -> CWLogEvent:
        timestamp = self._colored.yellow(timestamp_to_iso(int(event.timestamp)))
        log_stream_name = self._colored.cyan(event.log_stream_name)
        event.message = f"{log_stream_name} {timestamp} {event.message}"
        return event


class CWAddNewLineIfItDoesntExist(ObservabilityEventMapper):
    """
    Mapper implementation which will add new lines at the end of events if it is not already there
    """

    def map(self, event: Any) -> Any:
        # if it is a CWLogEvent, append new line at the end of event.message
        if isinstance(event, CWLogEvent) and not event.message.endswith("\n"):
            event.message = f"{event.message}\n"
            return event
        # if event is a string, then append new line at the end of the string
        if isinstance(event, str) and not event.endswith("\n"):
            return f"{event}\n"
        # no-action for unknown events
        return event


class CWLogEventJSONMapper(ObservabilityEventMapper[CWLogEvent]):
    """
    Converts given CWLogEvent into JSON string
    """

    # pylint: disable=no-self-use
    def map(self, event: CWLogEvent) -> CWLogEvent:
        event.message = json.dumps(event.event)
        return event



"""
Class to publish metrics
"""

import logging

import requests

# Get the preconfigured endpoint URL
from samcli.cli.global_config import GlobalConfig
from samcli.settings import telemetry_endpoint_url as DEFAULT_ENDPOINT_URL

LOG = logging.getLogger(__name__)


class Telemetry:
    def __init__(self, url=None):
        """
        Initialize the Telemetry object.

        Parameters
        ----------
        url : str
            Optional, URL where the metrics should be published to
        """
        self._url = url or DEFAULT_ENDPOINT_URL
        LOG.debug("Telemetry endpoint configured to be %s", self._url)

    def emit(self, metric, force_emit=False):
        """
        Emits the metric with given name and the attributes and send it immediately to the HTTP backend. This method
        will return immediately without waiting for response from the backend. Before sending, this method will
        also update ``attrs`` with some common attributes used by all metrics.

        Parameters
        ----------
        metric : Metric
            Metric to be published

        force_emit : bool
            Defaults to False. Set to True to emit even when telemetry is turned off.
        """
        if bool(GlobalConfig().telemetry_enabled) or force_emit:
            self._send({metric.get_metric_name(): metric.get_data()})

    def _send(self, metric, wait_for_response=False):
        """
        Serializes the metric data to JSON and sends to the backend.

        Parameters
        ----------

        metric : dict
            Dictionary of metric data to send to backend.

        wait_for_response : bool
            If set to True, this method will wait until the HTTP server returns a response. If not, it will return
            immediately after the request is sent.
        """

        if not self._url:
            # Endpoint not configured. So simply return
            LOG.debug("Not sending telemetry. Endpoint URL not configured")
            return

        payload = {"metrics": [metric]}
        LOG.debug("Sending Telemetry: %s", payload)

        timeout_ms = 2000 if wait_for_response else 100  # 2 seconds to wait for response or 100ms

        timeout = (
            2,  # connection timeout. Always set to 2 seconds
            timeout_ms / 1000.0,  # Read timeout. Tweaked based on input.
        )
        try:
            r = requests.post(self._url, json=payload, timeout=timeout)
            LOG.debug("Telemetry response: %d", r.status_code)
        except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as ex:
            # Expected if request times out OR cannot connect to the backend (offline).
            # Just print debug log and ignore the exception.
            LOG.debug(str(ex))



"""
Supplies the environment variables necessary to set up Local Lambda runtime
"""

import sys
from enum import IntEnum


class Python(IntEnum):
    TWO = 2


class EnvironmentVariables:
    """
    Use this class to get the environment variables necessary to run the Lambda function. It returns the AWS specific
    variables (credentials, regions, etc) along with any environment variables configured on the function.

    Customers define the name of the environment variables along with default values, if any, when creating the
    function. In order to test the function with different scenarios, customers can override values for some of the
    variables. This class supports three mechanisms of providing values to environment variables.
    If a variable is given a value through all the three mechanisms, then the value from higher priority will be used:

    Priority (Highest to Lowest):
        - Override Values - User specified these
        - Shell Environment Values - Came from the shell environment
        - Default Values - Hard coded values

    If a variable does *not* get a value from either of the above mechanisms, it is given a value of "" (empty string).
    If the value of a variable is an intrinsic function dict/list, then it is given a value of "" (empty string).

    If real AWS Credentials were supplied, this class will expose them through appropriate environment variables.
    If not, this class will provide the following placeholder values to AWS Credentials:
        region = "us-east-1"
        key = "defaultkey"
        secret = "defaultsecret"
    """

    _BLANK_VALUE = ""
    _DEFAULT_AWS_CREDS = {"region": "us-east-1", "key": "defaultkey", "secret": "defaultsecret"}

    def __init__(
        self,
        function_name=None,
        function_memory=None,
        function_timeout=None,
        function_handler=None,
        variables=None,
        shell_env_values=None,
        override_values=None,
        aws_creds=None,
    ):
        """
        Initializes this class. It takes in two sets of properties:
            a) (Required) Function information
            b) (Optional) Environment variable configured on the function

        :param str function_name: The name of the function
        :param integer function_memory: Memory size of the function in megabytes
        :param integer function_timeout: Function's timeout in seconds
        :param string function_handler: Handler of the function
        :param dict variables: Optional. Dict whose key is the environment variable names and value is the default
            values for the variable.
        :param dict shell_env_values: Optional. Dict containing values for the variables grabbed from the shell's
            environment.
        :param dict override_values: Optional. Dict containing values for the variables that will override the values
            from ``default_values`` and ``shell_env_values``.
        :param dict aws_creds: Optional. Dictionary containing AWS credentials passed to the Lambda runtime through
            environment variables. It should contain "key", "secret", "region" and optional "sessiontoken" keys
        """

        self._function = {
            "memory": function_memory,
            "timeout": function_timeout,
            "handler": function_handler,
            "name": function_name,
        }

        self.variables = variables or {}
        self.shell_env_values = shell_env_values or {}
        self.override_values = override_values or {}
        self.aws_creds = aws_creds or {}

    def resolve(self):
        """
        Resolves the values from different sources and returns a dict of environment variables to use when running
        the function locally.

        :return dict: Dict where key is the variable name and value is the value of the variable. Both key and values
            are strings
        """

        # AWS_* variables must always be passed to the function, but user has the choice to override them
        result = self._get_aws_variables()

        # Default value for the variable gets lowest priority
        for name, value in self.variables.items():

            # Shell environment values, second priority
            if name in self.shell_env_values:
                value = self.shell_env_values[name]

            # Overridden values, highest priority
            if name in self.override_values:
                value = self.override_values[name]

            # Any value must be a string when passed to Lambda runtime.
            # Runtime expects a Map<String, String> for environment variables
            result[name] = self._stringify_value(value)

        return result

    def add_lambda_event_body(self, value):
        """
        Adds the value of AWS_LAMBDA_EVENT_BODY environment variable.
        """
        self.variables["AWS_LAMBDA_EVENT_BODY"] = value

    @property
    def timeout(self):
        return self._function["timeout"]

    @timeout.setter
    def timeout(self, value):
        self._function["timeout"] = value

    @property
    def memory(self):
        return self._function["memory"]

    @memory.setter
    def memory(self, value):
        self._function["memory"] = value

    @property
    def handler(self):
        return self._function["handler"]

    @handler.setter
    def handler(self, value):
        self._function["handler"] = value

    @property
    def name(self):
        return self._function["name"]

    @name.setter
    def name(self, value):
        self._function["name"] = value

    def _get_aws_variables(self):
        """
        Returns the AWS specific environment variables that should be available in the Lambda runtime.
        They are prefixed it "AWS_*".

        :return dict: Name and value of AWS environment variable
        """

        result = {
            # Variable that says this function is running in Local Lambda
            "AWS_SAM_LOCAL": "true",
            # Function configuration
            "AWS_LAMBDA_FUNCTION_MEMORY_SIZE": str(self.memory),
            "AWS_LAMBDA_FUNCTION_TIMEOUT": str(self.timeout),
            "AWS_LAMBDA_FUNCTION_HANDLER": self._function["handler"],
            "AWS_LAMBDA_FUNCTION_NAME": str(self.name),
            "AWS_LAMBDA_FUNCTION_VERSION": "$LATEST",
            "AWS_LAMBDA_LOG_GROUP_NAME": f"aws/lambda/{self.name}",
            "AWS_LAMBDA_LOG_STREAM_NAME": "$LATEST",
            # AWS Credentials - Use the input credentials or use the defaults
            "AWS_REGION": self.aws_creds.get("region", self._DEFAULT_AWS_CREDS["region"]),
            "AWS_DEFAULT_REGION": self.aws_creds.get("region", self._DEFAULT_AWS_CREDS["region"]),
            "AWS_ACCESS_KEY_ID": self.aws_creds.get("key", self._DEFAULT_AWS_CREDS["key"]),
            "AWS_SECRET_ACCESS_KEY": self.aws_creds.get("secret", self._DEFAULT_AWS_CREDS["secret"]),
            "AWS_ACCOUNT_ID": "123456789012",
        }

        # Session Token should be added **only** if the input creds have a token and the value is not empty.
        if self.aws_creds.get("sessiontoken"):
            result["AWS_SESSION_TOKEN"] = self.aws_creds.get("sessiontoken")

        return result

    def _stringify_value(self, value):
        """
        This method stringifies values of environment variables. If the value of the method is a list or dictionary,
        then this method will replace it with empty string. Values of environment variables in Lambda must be a string.
        List or dictionary usually means they are intrinsic functions which have not been resolved.

        :param value: Value to stringify
        :return string: Stringified value
        """

        # List/dict/None values are replaced with a blank
        if isinstance(value, (dict, list, tuple)) or value is None:
            result = self._BLANK_VALUE

        # str(True) will output "True". To maintain backwards compatibility we need to output "true" or "false"
        elif value is True:
            result = "true"
        elif value is False:  # pylint: disable=compare-to-zero
            result = "false"

        # value is a scalar type like int, str which can be stringified
        # do not stringify unicode in Py2, Py3 str supports unicode
        elif sys.version_info.major > Python.TWO:
            result = str(value)
        elif not isinstance(value, unicode):  # noqa: F821 pylint: disable=undefined-variable
            result = str(value)
        else:
            result = value

        return result

    def __eq__(self, other):
        if not isinstance(other, EnvironmentVariables):
            return False
        return self.resolve() == other.resolve()



"""
Library for Validating Sam Templates
"""
import functools
import logging

from boto3.session import Session
from samtranslator.parser import parser
from samtranslator.public.exceptions import InvalidDocumentException
from samtranslator.translator.translator import Translator

from samcli.commands.validate.lib.exceptions import InvalidSamDocumentException
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.lib.utils.resources import AWS_SERVERLESS_FUNCTION
from samcli.yamlhelper import yaml_dump

LOG = logging.getLogger(__name__)


class SamTemplateValidator:
    def __init__(self, sam_template, managed_policy_loader, profile=None, region=None):
        """
        Construct a SamTemplateValidator

        Design Details:

        managed_policy_loader is injected into the `__init__` to allow future expansion
        and overriding capabilities. A typically pattern is to pass the name of the class into
        the `__init__` as keyword args. As long as the class 'conforms' to the same 'interface'.
        This allows the class to be changed by the client and allowing customization of the class being
        initialized. Something I had in mind would be allowing a template to be run and checked
        'offline' (not needing aws creds). To make this an easier transition in the future, we ingest
        the ManagedPolicyLoader class.

        Parameters
        ----------
        sam_template dict
            Dictionary representing a SAM Template
        managed_policy_loader ManagedPolicyLoader
            Sam ManagedPolicyLoader
        """
        self.sam_template = sam_template
        self.managed_policy_loader = managed_policy_loader
        self.sam_parser = parser.Parser()
        self.boto3_session = Session(profile_name=profile, region_name=region)

    def get_translated_template_if_valid(self):
        """
        Runs the SAM Translator to determine if the template provided is valid. This is similar to running a
        ChangeSet in CloudFormation for a SAM Template

        Raises
        -------
        InvalidSamDocumentException
             If the template is not valid, an InvalidSamDocumentException is raised
        """
        managed_policy_map = self.managed_policy_loader.load()

        sam_translator = Translator(
            managed_policy_map=managed_policy_map,
            sam_parser=self.sam_parser,
            plugins=[],
            boto_session=self.boto3_session,
        )

        self._replace_local_codeuri()
        self._replace_local_image()

        try:
            template = sam_translator.translate(sam_template=self.sam_template, parameter_values={})
            LOG.debug("Translated template is:\n%s", yaml_dump(template))
            return yaml_dump(template)
        except InvalidDocumentException as e:
            raise InvalidSamDocumentException(
                functools.reduce(lambda message, error: message + " " + str(error), e.causes, str(e))
            ) from e

    def _replace_local_codeuri(self):
        """
        Replaces the CodeUri in AWS::Serverless::Function and DefinitionUri in AWS::Serverless::Api and
        AWS::Serverless::HttpApi to a fake S3 Uri. This is to support running the SAM Translator with
        valid values for these fields. If this in not done, the template is invalid in the eyes of SAM
        Translator (the translator does not support local paths)
        """

        all_resources = self.sam_template.get("Resources", {})
        global_settings = self.sam_template.get("Globals", {})

        for resource_type, properties in global_settings.items():

            if resource_type == "Function":
                if all(
                    [
                        _properties.get("Properties", {}).get("PackageType", ZIP) == ZIP
                        for _, _properties in all_resources.items()
                    ]
                    + [_properties.get("PackageType", ZIP) == ZIP for _, _properties in global_settings.items()]
                ):
                    SamTemplateValidator._update_to_s3_uri("CodeUri", properties)

        for _, resource in all_resources.items():

            resource_type = resource.get("Type")
            resource_dict = resource.get("Properties", {})

            if resource_type == "AWS::Serverless::Function" and resource_dict.get("PackageType", ZIP) == ZIP:

                SamTemplateValidator._update_to_s3_uri("CodeUri", resource_dict)

            if resource_type == "AWS::Serverless::LayerVersion":

                SamTemplateValidator._update_to_s3_uri("ContentUri", resource_dict)

            if resource_type == "AWS::Serverless::Api":
                if "DefinitionUri" in resource_dict:
                    SamTemplateValidator._update_to_s3_uri("DefinitionUri", resource_dict)

            if resource_type == "AWS::Serverless::HttpApi":
                if "DefinitionUri" in resource_dict:
                    SamTemplateValidator._update_to_s3_uri("DefinitionUri", resource_dict)

            if resource_type == "AWS::Serverless::StateMachine":
                if "DefinitionUri" in resource_dict:
                    SamTemplateValidator._update_to_s3_uri("DefinitionUri", resource_dict)

    def _replace_local_image(self):
        """
        Adds fake ImageUri to AWS::Serverless::Functions that reference a local image using Metadata.
        This ensures sam validate works without having to package the app or use ImageUri.
        """
        resources = self.sam_template.get("Resources", {})
        for _, resource in resources.items():
            resource_type = resource.get("Type")
            properties = resource.get("Properties", {})

            is_image_function = resource_type == AWS_SERVERLESS_FUNCTION and properties.get("PackageType") == IMAGE
            is_local_image = resource.get("Metadata", {}).get("Dockerfile")

            if is_image_function and is_local_image:
                if "ImageUri" not in properties:
                    properties["ImageUri"] = "111111111111.dkr.ecr.region.amazonaws.com/repository"

    @staticmethod
    def is_s3_uri(uri):
        """
        Checks the uri and determines if it is a valid S3 Uri

        Parameters
        ----------
        uri str, required
            Uri to check

        Returns
        -------
        bool
            Returns True if the uri given is an S3 uri, otherwise False

        """
        return isinstance(uri, str) and uri.startswith("s3://")

    @staticmethod
    def _update_to_s3_uri(property_key, resource_property_dict, s3_uri_value="s3://bucket/value"):
        """
        Updates the 'property_key' in the 'resource_property_dict' to the value of 's3_uri_value'

        Note: The function will mutate the resource_property_dict that is pass in

        Parameters
        ----------
        property_key str, required
            Key in the resource_property_dict
        resource_property_dict dict, required
            Property dictionary of a Resource in the template to replace
        s3_uri_value str, optional
            Value to update the value of the property_key to
        """
        uri_property = resource_property_dict.get(property_key, ".")

        # ignore if dict or already an S3 Uri
        if isinstance(uri_property, dict) or SamTemplateValidator.is_s3_uri(uri_property):
            return

        resource_property_dict[property_key] = s3_uri_value



"""
File keeps Factory method to prepare required puller information
with its producers and consumers
"""
import logging
from typing import List, Optional

from botocore.exceptions import ClientError

from samcli.commands.exceptions import UserException
from samcli.commands.logs.console_consumers import CWConsoleEventConsumer
from samcli.commands.traces.traces_puller_factory import generate_trace_puller
from samcli.lib.observability.cw_logs.cw_log_formatters import (
    CWAddNewLineIfItDoesntExist,
    CWColorizeErrorsFormatter,
    CWJsonFormatter,
    CWKeywordHighlighterFormatter,
    CWLogEventJSONMapper,
    CWPrettyPrintFormatter,
)
from samcli.lib.observability.cw_logs.cw_log_group_provider import LogGroupProvider
from samcli.lib.observability.cw_logs.cw_log_puller import CWLogPuller
from samcli.lib.observability.observability_info_puller import (
    ObservabilityCombinedPuller,
    ObservabilityEventConsumer,
    ObservabilityEventConsumerDecorator,
    ObservabilityPuller,
)
from samcli.lib.observability.util import OutputOption
from samcli.lib.utils.boto_utils import BotoProviderType, get_client_error_code
from samcli.lib.utils.cloudformation import CloudFormationResourceSummary
from samcli.lib.utils.colors import Colored

LOG = logging.getLogger(__name__)


class NoPullerGeneratedException(UserException):
    """
    Used to indicate that no puller information have been generated
    therefore there is no observability information (logs, xray) to pull
    """


def generate_puller(
    boto_client_provider: BotoProviderType,
    resource_information_list: List[CloudFormationResourceSummary],
    filter_pattern: Optional[str] = None,
    additional_cw_log_groups: Optional[List[str]] = None,
    output: OutputOption = OutputOption.text,
    include_tracing: bool = False,
) -> ObservabilityPuller:
    """
    This function will generate generic puller which can be used to
    pull information from various observability resources.

    Parameters
    ----------
    boto_client_provider: BotoProviderType
        Boto3 client generator, which will create a new instance of the client with a new session that could be
        used within different threads/coroutines
    resource_information_list : List[CloudFormationResourceSummary]
        List of resource information, which keeps logical id, physical id and type of the resources
    filter_pattern : Optional[str]
        Optional filter pattern which will be used to filter incoming events
    additional_cw_log_groups : Optional[str]
        Optional list of additional CloudWatch log groups which will be used to fetch
        log events from.
    output : OutputOption
        Decides how the output will be presented in the console. It is been used to select correct consumer type
        between (default) text consumer or json consumer
    include_tracing: bool
        A flag to include the xray traces log or not

    Returns
    -------
        Puller instance that can be used to pull information.
    """
    if additional_cw_log_groups is None:
        additional_cw_log_groups = []
    pullers: List[ObservabilityPuller] = []

    # populate all puller instances for given resources
    for resource_information in resource_information_list:
        cw_log_group_name = LogGroupProvider.for_resource(
            boto_client_provider,
            resource_information.resource_type,
            resource_information.physical_resource_id,
        )
        if not cw_log_group_name:
            LOG.debug("Can't find CloudWatch LogGroup name for resource (%s)", resource_information.logical_resource_id)
            continue

        consumer = generate_consumer(filter_pattern, output, resource_information.logical_resource_id)
        pullers.append(
            CWLogPuller(
                boto_client_provider("logs"),
                consumer,
                cw_log_group_name,
                resource_information.logical_resource_id,
            )
        )

    # populate puller instances for the additional CloudWatch log groups
    for cw_log_group in additional_cw_log_groups:
        consumer = generate_consumer(filter_pattern, output)
        logs_client = boto_client_provider("logs")
        _validate_cw_log_group_name(cw_log_group, logs_client)
        pullers.append(
            CWLogPuller(
                logs_client,
                consumer,
                cw_log_group,
            )
        )

    # if tracing flag is set, add the xray traces puller to fetch debug traces
    if include_tracing:
        trace_puller = generate_trace_puller(boto_client_provider("xray"), output)
        pullers.append(trace_puller)

    # if no puller have been collected, raise an exception since there is nothing to pull
    if not pullers:
        raise NoPullerGeneratedException("No valid resources find to pull information")

    # return the combined puller instance, which will pull from all pullers collected
    return ObservabilityCombinedPuller(pullers)


def _validate_cw_log_group_name(cw_log_group, logs_client):
    try:
        _ = logs_client.describe_log_streams(logGroupName=cw_log_group, limit=1)
    except ClientError as ex:
        if get_client_error_code(ex) == "ResourceNotFoundException":
            LOG.warning("CloudWatch log group name (%s) does not exist.", cw_log_group)


def generate_consumer(
    filter_pattern: Optional[str] = None, output: OutputOption = OutputOption.text, resource_name: Optional[str] = None
):
    """
    Generates consumer instance with the given variables.
    If output is JSON, then it will return consumer with formatters for just JSON.
    Otherwise, it will return regular text console consumer
    """
    if output == OutputOption.json:
        return generate_json_consumer()

    return generate_text_consumer(filter_pattern)


def generate_json_consumer() -> ObservabilityEventConsumer:
    """
    Creates event consumer, which prints CW Log Events as JSON into terminal

    Returns
    -------
        ObservabilityEventConsumer which will store events into a file
    """
    return ObservabilityEventConsumerDecorator(
        [
            CWLogEventJSONMapper(),
        ],
        CWConsoleEventConsumer(True),
    )


def generate_text_consumer(filter_pattern: Optional[str]) -> ObservabilityEventConsumer:
    """
    Creates a console event consumer, which is used to display events in the user's console

    Parameters
    ----------
    filter_pattern : str
        Filter pattern is used to display certain words in a different pattern then
        the rest of the messages.

    Returns
    -------
        A consumer which will display events into console
    """
    colored = Colored()
    return ObservabilityEventConsumerDecorator(
        [
            CWColorizeErrorsFormatter(colored),
            CWJsonFormatter(),
            CWKeywordHighlighterFormatter(colored, filter_pattern),
            CWPrettyPrintFormatter(colored),
            CWAddNewLineIfItDoesntExist(),
        ],
        CWConsoleEventConsumer(),
    )



"""SyncFlow interface for HttpApi and RestApi"""
import logging
from abc import ABC
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from samcli.lib.providers.provider import ResourceIdentifier, Stack, get_resource_by_id
from samcli.lib.sync.sync_flow import ResourceAPICall, SyncFlow, get_definition_path
from samcli.lib.utils.hash import str_checksum

# BuildContext and DeployContext will only be imported for type checking to improve performance
# since no istances of contexts will be instantiated in this class
if TYPE_CHECKING:  # pragma: no cover
    from samcli.commands.build.build_context import BuildContext
    from samcli.commands.deploy.deploy_context import DeployContext
    from samcli.commands.sync.sync_context import SyncContext

LOG = logging.getLogger(__name__)


class GenericApiSyncFlow(SyncFlow, ABC):
    """SyncFlow interface for HttpApi and RestApi"""

    _api_client: Any
    _api_identifier: str
    _definition_uri: Optional[Path]
    _stacks: List[Stack]
    _swagger_body: Optional[bytes]

    def __init__(
        self,
        api_identifier: str,
        build_context: "BuildContext",
        deploy_context: "DeployContext",
        sync_context: "SyncContext",
        physical_id_mapping: Dict[str, str],
        log_name: str,
        stacks: List[Stack],
    ):
        """
        Parameters
        ----------
        api_identifier : str
            HttpApi resource identifier that needs to have associated Api updated.
        build_context : BuildContext
            BuildContext used for build related parameters
        deploy_context : BuildContext
            DeployContext used for this deploy related parameters
        sync_context: SyncContext
            SyncContext object that obtains sync information.
        physical_id_mapping : Dict[str, str]
            Mapping between resource logical identifier and physical identifier
        log_name: str
            Log name passed from subclasses, HttpApi or RestApi
        stacks : List[Stack], optional
            List of stacks containing a root stack and optional nested stacks
        """
        super().__init__(
            build_context,
            deploy_context,
            sync_context,
            physical_id_mapping,
            log_name=log_name,
            stacks=stacks,
        )
        self._api_identifier = api_identifier

    @property
    def sync_state_identifier(self) -> str:
        """
        Sync state is the unique identifier for each sync flow
        In sync state toml file we will store
        Key as HttpApiSyncFlow:HttpApiLogicalId or RestApiSyncFlow:RestApiLogicalId
        Value as API definition hash
        """
        return self.__class__.__name__ + ":" + self._api_identifier

    def gather_resources(self) -> None:
        self._definition_uri = self._get_definition_file(self._api_identifier)
        self._swagger_body = self._process_definition_file()
        if self._swagger_body:
            self._local_sha = str_checksum(self._swagger_body.decode("utf-8"))

    def _process_definition_file(self) -> Optional[bytes]:
        if self._definition_uri is None:
            return None
        with open(str(self._definition_uri), "rb") as swagger_file:
            swagger_body = swagger_file.read()
            return swagger_body

    def _get_definition_file(self, api_identifier: str) -> Optional[Path]:
        api_resource = get_resource_by_id(self._stacks, ResourceIdentifier(api_identifier))
        if not api_resource:
            return None
        return get_definition_path(
            api_resource,
            self._api_identifier,
            self._build_context.use_base_dir,
            self._build_context.base_dir,
            self._stacks,
        )

    def compare_remote(self) -> bool:
        return False

    def gather_dependencies(self) -> List[SyncFlow]:
        return []

    def _get_resource_api_calls(self) -> List[ResourceAPICall]:
        return []

    def _equality_keys(self) -> Any:
        return self._api_identifier



"""
Provide a CDK implementation of IaCPluginInterface
"""
from typing import List

from samcli.lib.iac.plugins_interfaces import IaCPluginInterface, LookupPath, SamCliProject, Stack


# TODO: Implement the new interface methods for the CDK plugin type
class CdkIacImplementation(IaCPluginInterface):
    """
    CDK implementation for the plugins interface.
    read_project parses the CDK and returns a SamCliProject object
    write_project writes the updated project
        back to the build dir and returns true if successful
    update_packaged_locations updates the package locations and r
        returns true if successful
    get_iac_file_types returns a list of file types/patterns associated with
        the CDK project type
    """

    def read_project(self, lookup_paths: List[LookupPath]) -> SamCliProject:
        pass

    def write_project(self, project: SamCliProject, build_dir: str) -> bool:
        pass

    def update_packaged_locations(self, stack: Stack) -> bool:
        pass

    @staticmethod
    def get_iac_file_patterns() -> List[str]:
        return ["cdk.json"]



"""
This file contains common CLI options common to all commands. As we add more commands, this will
become a repository of options that other commands could use when needed.
"""

import click

from .context import Context


def debug_option(f):
    """
    Configures --debug option for CLI

    :param f: Callback Function to be passed to Click
    """

    def callback(ctx, param, value):
        state = ctx.ensure_object(Context)
        state.debug = value
        return value

    return click.option(
        "--debug",
        expose_value=False,
        is_eager=True,
        is_flag=True,
        envvar="SAM_DEBUG",
        help="Turn on debug logging to print debug message generated by AWS SAM CLI and display timestamps.",
        callback=callback,
    )(f)


def region_option(f):
    """
    Configures --region option for CLI

    :param f: Callback Function to be passed to Click
    """

    def callback(ctx, param, value):
        state = ctx.ensure_object(Context)
        from botocore import exceptions, utils

        from samcli.commands.exceptions import RegionError

        try:
            utils.validate_region_name(value)
        except exceptions.InvalidRegionError as ex:
            raise RegionError(
                message=f"Provided region: {value} doesn't match a supported format", wrapped_from=ex.__class__.__name__
            ) from ex
        state.region = value
        return value

    return click.option(
        "--region", expose_value=False, help="Set the AWS Region of the service (e.g. us-east-1).", callback=callback
    )(f)


def profile_option(f):
    """
    Configures --profile option for CLI

    :param f: Callback Function to be passed to Click
    """

    def callback(ctx, param, value):
        state = ctx.ensure_object(Context)
        state.profile = value
        return value

    return click.option(
        "--profile",
        expose_value=False,
        help="Select a specific profile from your credential file to get AWS credentials.",
        callback=callback,
    )(f)



"""
Represents Events and their values.
"""

import logging
import threading
from datetime import datetime
from enum import Enum
from typing import List, Optional

from samcli.cli.context import Context
from samcli.lib.build.workflows import ALL_CONFIGS
from samcli.lib.telemetry.telemetry import Telemetry
from samcli.local.common.runtime_template import INIT_RUNTIMES

LOG = logging.getLogger(__name__)


class EventName(Enum):
    """Enum for the names of available events to track."""

    USED_FEATURE = "UsedFeature"
    BUILD_FUNCTION_RUNTIME = "BuildFunctionRuntime"
    SYNC_USED = "SyncUsed"
    SYNC_FLOW_START = "SyncFlowStart"
    SYNC_FLOW_END = "SyncFlowEnd"
    BUILD_WORKFLOW_USED = "BuildWorkflowUsed"


class UsedFeature(Enum):
    """Enum for the names of event values of UsedFeature"""

    ACCELERATE = "Accelerate"
    CDK = "CDK"
    INIT_WITH_APPLICATION_INSIGHTS = "InitWithApplicationInsights"
    CFNLint = "CFNLint"


class EventType:
    """Class for Events and the types of values they may have."""

    _SYNC_FLOWS = [
        "AliasVersionSyncFlow",
        "AutoDependencyLayerSyncFlow",
        "AutoDependencyLayerParentSyncFlow",
        "FunctionSyncFlow",
        "FunctionLayerReferenceSync",
        "GenericApiSyncFlow",
        "HttpApiSyncFlow",
        "ImageFunctionSyncFlow",
        "LayerSyncFlow",
        "RestApiSyncFlow",
        "StepFunctionsSyncFlow",
        "ZipFunctionSyncFlow",
    ]
    _WORKFLOWS = [f"{config.language}-{config.dependency_manager}" for config in ALL_CONFIGS]

    _event_values = {  # Contains allowable values for Events
        EventName.USED_FEATURE: [event.value for event in UsedFeature],
        EventName.BUILD_FUNCTION_RUNTIME: INIT_RUNTIMES,
        EventName.SYNC_USED: [
            "Start",
            "End",
        ],
        EventName.SYNC_FLOW_START: _SYNC_FLOWS,
        EventName.SYNC_FLOW_END: _SYNC_FLOWS,
        EventName.BUILD_WORKFLOW_USED: _WORKFLOWS,
    }

    @staticmethod
    def get_accepted_values(event_name: EventName) -> List[str]:
        """Get all acceptable values for a given Event name."""
        if event_name not in EventType._event_values:
            return []
        return EventType._event_values[event_name]


class Event:
    """Class to represent Events that occur in SAM CLI."""

    event_name: EventName
    event_value: str  # Validated by EventType.get_accepted_values to never be an arbitrary string
    thread_id = threading.get_ident()  # The thread ID; used to group Events from the same command run
    time_stamp: str

    def __init__(self, event_name: str, event_value: str):
        Event._verify_event(event_name, event_value)
        self.event_name = EventName(event_name)
        self.event_value = event_value
        self.time_stamp = str(datetime.utcnow())[:-3]  # format microseconds from 6 -> 3 figures to allow SQL casting

    def __eq__(self, other):
        return self.event_name == other.event_name and self.event_value == other.event_value

    def __repr__(self):
        return (
            f"Event(event_name={self.event_name.value}, "
            f"event_value={self.event_value}, "
            f"thread_id={self.thread_id}, "
            f"time_stamp={self.time_stamp})"
        )

    def to_json(self):
        return {
            "event_name": self.event_name.value,
            "event_value": self.event_value,
            "thread_id": self.thread_id,
            "time_stamp": self.time_stamp,
        }

    @staticmethod
    def _verify_event(event_name: str, event_value: str) -> None:
        """Raise an EventCreationError if either the event name or value is not valid."""
        if event_name not in Event._get_event_names():
            raise EventCreationError(f"Event '{event_name}' does not exist.")
        if event_value not in EventType.get_accepted_values(EventName(event_name)):
            raise EventCreationError(f"Event '{event_name}' does not accept value '{event_value}'.")

    @staticmethod
    def _get_event_names() -> List[str]:
        """Retrieves a list of all valid event names."""
        return [event.value for event in EventName]


class EventTracker:
    """Class to track and recreate Events as they occur."""

    _events: List[Event] = []
    _event_lock = threading.Lock()
    _session_id: Optional[str] = None

    MAX_EVENTS: int = 50  # Maximum number of events to store before sending

    @staticmethod
    def track_event(event_name: str, event_value: str):
        """Method to track an event where and when it occurs.

        Place this method in the codepath of the event that you would
        like to track. For instance, if you would like to track when
        FeatureX is used, append this method to the end of that function.

        Parameters
        ----------
        event_name: str
            The name of the Event. Must be a valid EventName value, or an
            EventCreationError will be thrown.
        event_value: str
            The value of the Event. Must be a valid EventType value for the
            passed event_name, or an EventCreationError will be thrown.

        Examples
        --------
        >>> def feature_x(...):
                # do things
                EventTracker.track_event("UsedFeature", "FeatureX")

        >>> def feature_y(...) -> Any:
                # do things
                EventTracker.track_event("UsedFeature", "FeatureY")
                return some_value
        """
        try:
            should_send: bool = False
            with EventTracker._event_lock:
                EventTracker._events.append(Event(event_name, event_value))
                # Get the session ID (needed for multithreading sending)
                if not EventTracker._session_id:
                    try:
                        ctx = Context.get_current_context()
                        if ctx:
                            EventTracker._session_id = ctx.session_id
                    except RuntimeError:
                        LOG.debug("EventTracker: Unable to obtain session ID")
                if len(EventTracker._events) >= EventTracker.MAX_EVENTS:
                    should_send = True
            if should_send:
                EventTracker.send_events()
        except EventCreationError as e:
            LOG.debug("Error occurred while trying to track an event: %s", e)

    @staticmethod
    def get_tracked_events() -> List[Event]:
        """Retrieve a list of all currently tracked Events."""
        with EventTracker._event_lock:
            return EventTracker._events

    @staticmethod
    def clear_trackers():
        """Clear the current list of tracked Events before the next session."""
        with EventTracker._event_lock:
            EventTracker._events = []

    @staticmethod
    def send_events() -> threading.Thread:
        """Call a thread to send the current list of Events via Telemetry."""
        send_thread = threading.Thread(target=EventTracker._send_events_in_thread)
        send_thread.start()
        return send_thread

    @staticmethod
    def _send_events_in_thread():
        """Send the current list of Events via Telemetry."""
        from samcli.lib.telemetry.metric import Metric  # pylint: disable=cyclic-import

        msa = {}

        with EventTracker._event_lock:
            if not EventTracker._events:  # Don't do anything if there are no events to send
                return

            msa["events"] = [e.to_json() for e in EventTracker._events]
            EventTracker._events = []  # Manual clear_trackers() since we're within the lock

        telemetry = Telemetry()
        metric = Metric("events")
        metric.add_data("sessionId", EventTracker._session_id)
        metric.add_data("metricSpecificAttributes", msa)
        telemetry.emit(metric)


def track_long_event(start_event_name: str, start_event_value: str, end_event_name: str, end_event_value: str):
    """Decorator for tracking events that occur at start and end of a function.

    The decorator tracks two Events total, where the first Event occurs
    at the start of the decorated function's execution (prior to its
    first line) and the second Event occurs after the function has ended
    (after the final line of the function has executed).
    If this decorator is being placed in a function that also contains the
    `track_command` decorator, ensure that this decorator is placed BELOW
    `track_command`. Otherwise, the current list of Events will be sent
    before the end_event will be added, resulting in an additional 'events'
    metric with only that single Event.

    Parameters
    ----------
    start_event_name: str
        The name of the Event that is executed at the start of the
        decorated function's execution. Must be a valid EventName
        value or the decorator will not run.
    start_event_value: str
        The value of the Event that is executed at the start of the
        decorated function's execution. Must be a valid EventType
        value for the passed `start_event_name` or the decorator
        will not run.
    end_event_name: str
        The name of the Event that is executed at the end of the
        decorated function's execution. Must be a valid EventName
        value or the decorator will not run.
    end_event_value: str
        The value of the Event that is executed at the end of the
        decorated function's execution. Must be a valid EventType
        value for the passed `end_event_name` or the decorator
        will not run.

    Examples
    --------
    >>> @track_long_event("FuncStart", "Func1", "FuncEnd", "Func1")
        def func1(...):
            # do things

    >>> @track_long_event("FuncStart", "Func2", "FuncEnd", "Func2")
        def func2(...):
            # do things
    """
    should_track = True
    try:
        # Check that passed values are valid Events
        Event(start_event_name, start_event_value)
        Event(end_event_name, end_event_value)
    except EventCreationError as e:
        LOG.debug("Error occurred while trying to track an event: %s\nDecorator not run.", e)
        should_track = False

    def decorator_for_events(func):
        """The actual decorator"""

        def wrapped(*args, **kwargs):
            # Track starting event
            if should_track:
                EventTracker.track_event(start_event_name, start_event_value)
            exception = None
            # Run the function
            try:
                return_value = func(*args, **kwargs)
            except Exception as e:
                exception = e
            # Track ending event
            if should_track:
                EventTracker.track_event(end_event_name, end_event_value)
                EventTracker.send_events()  # Ensure Events are sent at the end of execution
            if exception:
                raise exception

            return return_value

        return wrapped

    return decorator_for_events


class EventCreationError(Exception):
    """Exception called when an Event is not properly created."""



"""LockDistributor for creating and managing a set of locks"""
import multiprocessing
import multiprocessing.managers
import threading
from enum import Enum, auto
from typing import Dict, Optional, Set, cast


class LockChain:
    """Wrapper class for acquiring multiple locks in the same order to prevent dead locks
    Can be used with `with` statement"""

    def __init__(self, lock_mapping: Dict[str, threading.Lock]):
        """
        Parameters
        ----------
        lock_mapping : Dict[str, threading.Lock]
            Dictionary of locks with keys being used as generating reproduciable order for aquiring and releasing locks.
        """
        self._locks = [value for _, value in sorted(lock_mapping.items())]

    def acquire(self) -> None:
        """Aquire all locks in the LockChain"""
        for lock in self._locks:
            lock.acquire()

    def release(self) -> None:
        """Release all locks in the LockChain"""
        for lock in self._locks:
            lock.release()

    def __enter__(self) -> "LockChain":
        self.acquire()
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        self.release()


class LockDistributorType(Enum):
    """Types of LockDistributor"""

    THREAD = auto()
    PROCESS = auto()


class LockDistributor:
    """Dynamic lock distributor that supports threads and processes.
    In the case of processes, both manager(server process) or shared memory can be used.
    """

    _lock_type: LockDistributorType
    _manager: Optional[multiprocessing.managers.SyncManager]
    _dict_lock: threading.Lock
    _locks: Dict[str, threading.Lock]

    def __init__(
        self,
        lock_type: LockDistributorType = LockDistributorType.THREAD,
        manager: Optional[multiprocessing.managers.SyncManager] = None,
    ):
        """[summary]

        Parameters
        ----------
        lock_type : LockDistributorType, optional
            Whether locking with threads or processes, by default LockDistributorType.THREAD
        manager : Optional[multiprocessing.managers.SyncManager], optional
            Optional process sync mananger for creating proxy locks, by default None
        """
        self._lock_type = lock_type
        self._manager = manager
        self._dict_lock = self._create_new_lock()
        self._locks = (
            self._manager.dict()
            if self._lock_type == LockDistributorType.PROCESS and self._manager is not None
            else dict()
        )

    def _create_new_lock(self) -> threading.Lock:
        """Create a new lock based on lock type

        Returns
        -------
        threading.Lock
            Newly created lock
        """
        if self._lock_type == LockDistributorType.THREAD:
            return threading.Lock()

        return self._manager.Lock() if self._manager is not None else cast(threading.Lock, multiprocessing.Lock())

    def get_lock(self, key: str) -> threading.Lock:
        """Retrieve a lock associating with the key
        If the lock does not exist, a new lock will be created.

        Parameters
        ----------
        key : Key for retrieving the lock

        Returns
        -------
        threading.Lock
            Lock associated with the key
        """
        with self._dict_lock:
            if key not in self._locks:
                self._locks[key] = self._create_new_lock()
            return self._locks[key]

    def get_locks(self, keys: Set[str]) -> Dict[str, threading.Lock]:
        """Retrieve a list of locks associating with keys

        Parameters
        ----------
        keys : Set[str]
            Set of keys for retrieving the locks

        Returns
        -------
        Dict[str, threading.Lock]
            Dictionary mapping keys to locks
        """
        lock_mapping = dict()
        for key in keys:
            lock_mapping[key] = self.get_lock(key)
        return lock_mapping

    def get_lock_chain(self, keys: Set[str]) -> LockChain:
        """Similar to get_locks, but retrieves a LockChain object instead of a dictionary

        Parameters
        ----------
        keys : Set[str]
            Set of keys for retrieving the locks

        Returns
        -------
        LockChain
            LockChain object containing all the locks associated with keys
        """
        return LockChain(self.get_locks(keys))



"""
An interactive flow that prompt the user for required information to bootstrap the AWS account of an environment
with the required infrastructure
"""
import os
import sys
from textwrap import dedent
from typing import Callable, List, Optional, Tuple
from xmlrpc.client import boolean

import click
from botocore.credentials import EnvProvider

from samcli.commands.exceptions import CredentialsError
from samcli.commands.pipeline.bootstrap.oidc_config import (
    BitbucketOidcConfig,
    GitHubOidcConfig,
    GitLabOidcConfig,
    OidcConfig,
)
from samcli.commands.pipeline.external_links import CONFIG_AWS_CRED_DOC_URL
from samcli.lib.bootstrap.bootstrap import get_current_account_id
from samcli.lib.utils.colors import Colored
from samcli.lib.utils.defaults import get_default_aws_region
from samcli.lib.utils.profile import list_available_profiles

GITHUB_ACTIONS = "github-actions"
GITLAB = "gitlab"
BITBUCKET = "bitbucket-pipelines"
OPEN_ID_CONNECT = "oidc"
OIDC_SUPPORTED_PROVIDER = [GITHUB_ACTIONS, GITLAB, BITBUCKET]
IAM = "iam"


class GuidedContext:

    SUPPORTED_OIDC_PROVIDERS = {"1": GITHUB_ACTIONS, "2": GITLAB, "3": BITBUCKET}
    OIDC_PROVIDER_NAME_MAPPINGS = {GITHUB_ACTIONS: "GitHub Actions", GITLAB: "GitLab", BITBUCKET: "Bitbucket"}
    # GitHub defaults: https://tinyurl.com/github-defaults
    # GitLab defaults: https://docs.gitlab.com/ee/ci/cloud_services/aws/#add-the-identity-provider
    DEFAULT_OIDC_URLS = {
        GITHUB_ACTIONS: "https://token.actions.githubusercontent.com",
        GITLAB: "https://gitlab.com",
        BITBUCKET: None,
    }
    DEFAULT_CLIENT_IDS = {GITHUB_ACTIONS: "sts.amazonaws.com", GITLAB: "https://gitlab.com", BITBUCKET: None}

    def __init__(
        self,
        oidc_config: OidcConfig,
        github_config: GitHubOidcConfig,
        gitlab_config: GitLabOidcConfig,
        bitbucket_config: BitbucketOidcConfig,
        profile: Optional[str] = None,
        stage_configuration_name: Optional[str] = None,
        pipeline_user_arn: Optional[str] = None,
        pipeline_execution_role_arn: Optional[str] = None,
        cloudformation_execution_role_arn: Optional[str] = None,
        artifacts_bucket_arn: Optional[str] = None,
        create_image_repository: bool = False,
        image_repository_arn: Optional[str] = None,
        region: Optional[str] = None,
        permissions_provider: Optional[str] = None,
        enable_oidc_option: boolean = True,
    ) -> None:
        self.profile = profile
        self.stage_configuration_name = stage_configuration_name
        self.pipeline_user_arn = pipeline_user_arn
        self.pipeline_execution_role_arn = pipeline_execution_role_arn
        self.cloudformation_execution_role_arn = cloudformation_execution_role_arn
        self.artifacts_bucket_arn = artifacts_bucket_arn
        self.create_image_repository = create_image_repository
        self.image_repository_arn = image_repository_arn
        self.region = region
        self.permissions_provider = permissions_provider
        self.oidc_config = oidc_config
        self.github_config = github_config
        self.gitlab_config = gitlab_config
        self.bitbucket_config = bitbucket_config
        self.enable_oidc_option = enable_oidc_option
        self.color = Colored()

    def _prompt_account_id(self) -> None:
        profiles = list_available_profiles()
        click.echo("The following AWS credential sources are available to use.")
        click.echo(
            dedent(
                f"""\
                To know more about configuration AWS credentials, visit the link below:
                {CONFIG_AWS_CRED_DOC_URL}\
                """
            )
        )
        has_env_creds = os.getenv(EnvProvider.ACCESS_KEY) and os.getenv(EnvProvider.SECRET_KEY)
        click.echo(f"\t1 - Environment variables{' (not available)' if not has_env_creds else ''}")
        for i, profile in enumerate(profiles):
            click.echo(f"\t{i + 2} - {profile} (named profile)")
        click.echo("\tq - Quit and configure AWS credentials")
        answer = click.prompt(
            "Select a credential source to associate with this stage",
            show_choices=False,
            show_default=False,
            type=click.Choice((["1"] if has_env_creds else []) + [str(i + 2) for i in range(len(profiles))] + ["q"]),
        )
        if answer == "q":
            sys.exit(0)
        elif answer == "1":
            # by default, env variable has higher precedence
            # https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-envvars.html#envvars-list
            self.profile = None
        else:
            self.profile = profiles[int(answer) - 2]

        try:
            account_id = get_current_account_id(self.profile)
            click.echo(
                self.color.green(f"Associated account {account_id} with configuration {self.stage_configuration_name}.")
            )
        except CredentialsError as ex:
            click.echo(f"{self.color.red(ex.message)}\n")
            self._prompt_account_id()

    def _prompt_stage_configuration_name(self) -> None:
        click.echo(
            "Enter a configuration name for this stage. "
            "This will be referenced later when you use the sam pipeline init command:"
        )
        self.stage_configuration_name = click.prompt(
            "Stage configuration name",
            default=self.stage_configuration_name,
            type=click.STRING,
        )

    def _prompt_region_name(self) -> None:
        self.region = click.prompt(
            "Enter the region in which you want these resources to be created",
            type=click.STRING,
            default=get_default_aws_region(),
        )

    def _prompt_pipeline_user(self) -> None:
        self.pipeline_user_arn = click.prompt(
            "Enter the pipeline IAM user ARN if you have previously created one, or we will create one for you",
            default="",
            type=click.STRING,
        )

    def _prompt_pipeline_execution_role(self) -> None:
        self.pipeline_execution_role_arn = click.prompt(
            "Enter the pipeline execution role ARN if you have previously created one, "
            "or we will create one for you",
            default="",
            type=click.STRING,
        )

    def _prompt_cloudformation_execution_role(self) -> None:
        self.cloudformation_execution_role_arn = click.prompt(
            "Enter the CloudFormation execution role ARN if you have previously created one, "
            "or we will create one for you",
            default="",
            type=click.STRING,
        )

    def _prompt_artifacts_bucket(self) -> None:
        self.artifacts_bucket_arn = click.prompt(
            "Please enter the artifact bucket ARN for your Lambda function. "
            "If you do not have a bucket, we will create one for you",
            default="",
            type=click.STRING,
        )

    def _prompt_image_repository(self) -> None:
        if click.confirm("Does your application contain any IMAGE type Lambda functions?"):
            self.image_repository_arn = click.prompt(
                "Please enter the ECR image repository ARN(s) for your Image type function(s)."
                "If you do not yet have a repository, we will create one for you",
                default="",
                type=click.STRING,
            )
            self.create_image_repository = not bool(self.image_repository_arn)
        else:
            self.create_image_repository = False

    def _prompt_permissions_provider(self) -> None:
        click.echo("Select a user permissions provider:")
        click.echo("\t1 - IAM (default)")
        click.echo("\t2 - OpenID Connect (OIDC)")
        user_provider = click.prompt("Choice", type=click.Choice((["1", "2"])), show_default=False, default="1")
        self.permissions_provider = OPEN_ID_CONNECT if user_provider == "2" else IAM

    def _prompt_oidc_provider(self) -> None:
        click.echo("Select an OIDC provider:")
        for (key, provider) in self.SUPPORTED_OIDC_PROVIDERS.items():
            click.echo("\t{key} - {provider}".format(key=key, provider=self.OIDC_PROVIDER_NAME_MAPPINGS[provider]))
        oidc_provider = click.prompt(
            "Choice",
            type=click.Choice((list(self.SUPPORTED_OIDC_PROVIDERS))),
            show_default=False,
        )
        self.oidc_config.oidc_provider = self.SUPPORTED_OIDC_PROVIDERS[oidc_provider]

    def _prompt_oidc_provider_url(self) -> None:
        self.oidc_config.oidc_provider_url = click.prompt(
            "Enter the URL of the OIDC provider",
            type=click.STRING,
            default=self.DEFAULT_OIDC_URLS[self.oidc_config.oidc_provider] if self.oidc_config.oidc_provider else None,
        )

    def _prompt_oidc_client_id(self) -> None:
        self.oidc_config.oidc_client_id = click.prompt(
            "Enter the OIDC client ID (sometimes called audience)",
            type=click.STRING,
            default=self.DEFAULT_CLIENT_IDS[self.oidc_config.oidc_provider] if self.oidc_config.oidc_provider else None,
        )

    def _prompt_subject_claim(self) -> None:
        if self.oidc_config.oidc_provider == GITHUB_ACTIONS:
            if not self.github_config.github_org:
                self._prompt_github_org()
            if not self.github_config.github_repo:
                self._prompt_github_repo()
            if not self.github_config.deployment_branch:
                self._prompt_deployment_branch()
        elif self.oidc_config.oidc_provider == GITLAB:
            if not self.gitlab_config.gitlab_group:
                self._prompt_gitlab_group()
            if not self.gitlab_config.gitlab_project:
                self._prompt_gitlab_project()
            if not self.gitlab_config.deployment_branch:
                self._prompt_deployment_branch()
        elif self.oidc_config.oidc_provider == BITBUCKET:
            if not self.bitbucket_config.bitbucket_repo_uuid:
                self._prompt_bitbucket_repo_uuid()

    def _prompt_bitbucket_repo_uuid(self) -> None:
        self.bitbucket_config.bitbucket_repo_uuid = click.prompt(
            "Enter the Bitbucket repository UUID", type=click.STRING
        )

    def _prompt_gitlab_group(self) -> None:
        self.gitlab_config.gitlab_group = click.prompt(
            "Enter the GitLab group that the code repository belongs to."
            " If there is no group enter your username instead",
            type=click.STRING,
        )

    def _prompt_gitlab_project(self) -> None:
        self.gitlab_config.gitlab_project = click.prompt("Enter GitLab project name", type=click.STRING)

    def _prompt_github_org(self) -> None:
        self.github_config.github_org = click.prompt(
            "Enter the GitHub organization that the code repository belongs to."
            " If there is no organization enter your username instead",
            type=click.STRING,
        )

    def _prompt_github_repo(self) -> None:
        self.github_config.github_repo = click.prompt("Enter GitHub repository name", type=click.STRING)

    def _prompt_deployment_branch(self) -> None:
        deployment_branch = click.prompt(
            "Enter the name of the branch that deployments will occur from", type=click.STRING, default="main"
        )
        self.github_config.deployment_branch = deployment_branch
        self.gitlab_config.deployment_branch = deployment_branch

    def _validate_oidc_provider_url(self) -> None:
        while not self.oidc_config.oidc_provider_url:
            click.echo("Please enter the URL of the OIDC provider")
            self._prompt_oidc_provider_url()
        while self.oidc_config.oidc_provider_url.find("https://") == -1:
            click.echo("Please ensure the OIDC URL begins with 'https://'")
            self._prompt_oidc_provider_url()

    def _get_user_inputs(self) -> List[Tuple[str, Callable[[], None]]]:
        inputs = [
            (f"Account: {get_current_account_id(self.profile)}", self._prompt_account_id),
            (f"Stage configuration name: {self.stage_configuration_name}", self._prompt_stage_configuration_name),
            (f"Region: {self.region}", self._prompt_region_name),
        ]

        if self.permissions_provider == OPEN_ID_CONNECT:
            inputs.extend(
                [
                    (
                        f"OIDC identity provider URL: {self.oidc_config.oidc_provider_url}",
                        self._prompt_oidc_provider_url,
                    ),
                    (f"OIDC client ID: {self.oidc_config.oidc_client_id}", self._prompt_oidc_client_id),
                ]
            )
            if self.oidc_config.oidc_provider == GITHUB_ACTIONS:
                inputs.extend(
                    [
                        (f"GitHub organization: {self.github_config.github_org}", self._prompt_github_org),
                        (f"GitHub repository: {self.github_config.github_repo}", self._prompt_github_repo),
                        (f"Deployment branch:  {self.github_config.deployment_branch}", self._prompt_deployment_branch),
                    ]
                )
            elif self.oidc_config.oidc_provider == GITLAB:
                inputs.extend(
                    [
                        (f"GitLab group: {self.gitlab_config.gitlab_group}", self._prompt_gitlab_group),
                        (f"GitLab project: {self.gitlab_config.gitlab_project}", self._prompt_gitlab_project),
                        (f"Deployment branch: {self.gitlab_config.deployment_branch}", self._prompt_deployment_branch),
                    ]
                )
            elif self.oidc_config.oidc_provider == BITBUCKET:
                inputs.extend(
                    [
                        (
                            f"Bitbucket repository UUID: {self.bitbucket_config.bitbucket_repo_uuid}",
                            self._prompt_bitbucket_repo_uuid,
                        ),
                    ]
                )
        else:
            inputs.extend(
                [
                    (
                        f"Pipeline user ARN: {self.pipeline_user_arn}"
                        if self.pipeline_user_arn
                        else "Pipeline user: [to be created]",
                        self._prompt_pipeline_user,
                    )
                ]
            )

        inputs.extend(
            [
                (
                    f"Pipeline execution role ARN: {self.pipeline_execution_role_arn}"
                    if self.pipeline_execution_role_arn
                    else "Pipeline execution role: [to be created]",
                    self._prompt_pipeline_execution_role,
                ),
                (
                    f"CloudFormation execution role ARN: {self.cloudformation_execution_role_arn}"
                    if self.cloudformation_execution_role_arn
                    else "CloudFormation execution role: [to be created]",
                    self._prompt_cloudformation_execution_role,
                ),
                (
                    f"Artifacts bucket ARN: {self.artifacts_bucket_arn}"
                    if self.artifacts_bucket_arn
                    else "Artifacts bucket: [to be created]",
                    self._prompt_artifacts_bucket,
                ),
                (
                    f"ECR image repository ARN: {self.image_repository_arn}"
                    if self.image_repository_arn
                    else f"ECR image repository: [{'to be created' if self.create_image_repository else 'skipped'}]",
                    self._prompt_image_repository,
                ),
            ]
        )
        return inputs

    def run(self) -> None:  # pylint: disable=too-many-branches
        """
        Runs an interactive questionnaire to prompt the user for the ARNs of the AWS resources(infrastructure) required
        for the pipeline to work. Users can provide all, none or some resources' ARNs and leave the remaining empty
        and it will be created by the bootstrap command
        """
        click.secho(self.color.bold("[1] Stage definition"))
        if self.stage_configuration_name:
            click.echo(f"Stage configuration name: {self.stage_configuration_name}")
        else:
            self._prompt_stage_configuration_name()
        click.echo()

        click.secho(self.color.bold("[2] Account details"))
        self._prompt_account_id()
        click.echo()

        if not self.region:
            self._prompt_region_name()

        if self.enable_oidc_option and not self.permissions_provider == OPEN_ID_CONNECT and not self.pipeline_user_arn:
            self._prompt_permissions_provider()

        if self.permissions_provider == OPEN_ID_CONNECT:
            if not self.oidc_config.oidc_provider:
                self._prompt_oidc_provider()
            if not self.oidc_config.oidc_provider_url:
                self._prompt_oidc_provider_url()
            self._validate_oidc_provider_url()
            if not self.oidc_config.oidc_client_id:
                self._prompt_oidc_client_id()
            self._prompt_subject_claim()
        elif self.pipeline_user_arn:
            click.echo(f"Pipeline IAM user ARN: {self.pipeline_user_arn}")
        else:
            self._prompt_pipeline_user()
        click.echo()

        click.secho(self.color.bold("[3] Reference application build resources"))

        if self.pipeline_execution_role_arn:
            click.echo(f"Pipeline execution role ARN: {self.pipeline_execution_role_arn}")
        else:
            self._prompt_pipeline_execution_role()

        if self.cloudformation_execution_role_arn:
            click.echo(f"CloudFormation execution role ARN: {self.cloudformation_execution_role_arn}")
        else:
            self._prompt_cloudformation_execution_role()

        if self.artifacts_bucket_arn:
            click.echo(f"Artifacts bucket ARN: {self.cloudformation_execution_role_arn}")
        else:
            self._prompt_artifacts_bucket()

        if self.image_repository_arn:
            click.echo(f"ECR image repository ARN: {self.image_repository_arn}")
        else:
            self._prompt_image_repository()
        click.echo()

        # Ask customers to confirm the inputs
        click.secho(self.color.bold("[4] Summary"))
        while True:
            inputs = self._get_user_inputs()
            click.secho("Below is the summary of the answers:")
            for i, (text, _) in enumerate(inputs):
                click.secho(f"\t{i + 1} - {text}")
            edit_input = click.prompt(
                text="Press enter to confirm the values above, or select an item to edit the value",
                default="0",
                show_choices=False,
                show_default=False,
                type=click.Choice(["0"] + [str(i + 1) for i in range(len(inputs))]),
            )
            click.echo()
            if int(edit_input):
                inputs[int(edit_input) - 1][1]()
                click.echo()
            else:
                break



""" Isolates code download and merge logic for dynamic Schemas template """

import json
import os

import click
from botocore.exceptions import ClientError

from samcli.local.common.runtime_template import SAM_RUNTIME_TO_SCHEMAS_CODE_LANG_MAPPING
from samcli.local.lambdafn.zip import unzip


def do_download_source_code_binding(runtime, schema_template_details, schemas_api_caller, download_location):
    """
    Downloads source code binding for given registry and schema version,
    generating the code bindings if they haven't been generated first
    :param runtime: Lambda runtime
    :param schema_template_details: e.g: registry_name, schema_name, schema_version
    :param schemas_api_caller: the schemas api caller object
    :param download_location: the download location
    :return: directory location where code is downloaded
    """
    registry_name = schema_template_details["registry_name"]
    schema_name = schema_template_details["schema_full_name"]
    schema_version = schema_template_details["schema_version"]
    schemas_runtime = SAM_RUNTIME_TO_SCHEMAS_CODE_LANG_MAPPING.get(runtime)
    try:
        click.echo("Event code binding Registry: %s and Schema: %s" % (registry_name, schema_name))
        click.echo("Generating code bindings...")
        # Optimistically try to get the code bindings first...
        return schemas_api_caller.download_source_code_binding(
            schemas_runtime, registry_name, schema_name, schema_version, download_location
        )
    except ClientError as e:
        # If the code bindings are not available, the API throws a NotFoundException, so we need to generate them
        if e.response["Error"]["Code"] == "NotFoundException":
            # At this point, any exceptions should percolate out upwards, and be wrapped as generic SchemasApiException
            # put code binding (to trigger the code generation)
            schemas_api_caller.put_code_binding(schemas_runtime, registry_name, schema_name, schema_version)
            # wait till binding is generated
            schemas_api_caller.poll_for_code_binding_status(schemas_runtime, registry_name, schema_name, schema_version)
            # Try download code binding again
            return schemas_api_caller.download_source_code_binding(
                schemas_runtime, registry_name, schema_name, schema_version, download_location
            )
        raise e


def do_extract_and_merge_schemas_code(download_location, output_dir, project_name, template_location):
    """
    Unzips schemas generated code and merge it with cookiecutter genertaed source.
    :param download_location:
    :param output_dir:
    :param project_name:
    :param template_location:
    """
    click.echo("Merging code bindings...")
    cookiecutter_json_path = os.path.join(template_location, "cookiecutter.json")
    with open(cookiecutter_json_path, "r") as cookiecutter_json:
        cookiecutter_json_data = cookiecutter_json.read()
        cookiecutter_json = json.loads(cookiecutter_json_data)
        function_name = cookiecutter_json["function_name"]
        copy_location = os.path.join(output_dir, project_name, function_name)
        unzip(download_location, copy_location)



#!/usr/bin/env python

import io
import re
import os
from setuptools import setup, find_packages


def read(*filenames, **kwargs):
    encoding = kwargs.get("encoding", "utf-8")
    # io.open defaults to \n as universal line ending no matter on what system
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


def read_requirements(req="base.txt"):
    content = read(os.path.join("requirements", req))
    requirements = list()
    for line in content.split("\n"):
        line = line.strip()
        if line.startswith("#"):
            continue
        elif line.startswith("-r"):
            requirements.extend(read_requirements(line[3:]))
        else:
            requirements.append(line)
    return requirements


def read_version():
    content = read(os.path.join(os.path.dirname(__file__), "samcli", "__init__.py"))
    return re.search(r"__version__ = \"([^']+)\"", content).group(1)


cmd_name = "sam"
if os.getenv("SAM_CLI_DEV"):
    # We are installing in a dev environment
    cmd_name = "samdev"

setup(
    name="aws-sam-cli",
    version=read_version(),
    description="AWS SAM CLI is a CLI tool for local development and testing of Serverless applications",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Amazon Web Services",
    author_email="aws-sam-developers@amazon.com",
    url="https://github.com/aws/aws-sam-cli",
    license="Apache License 2.0",
    packages=find_packages(exclude=["tests.*", "tests", "installer.*", "installer"]),
    keywords="AWS SAM CLI",
    # Support Python 3.7 or greater
    python_requires=">=3.7, <=4.0, !=4.0",
    entry_points={"console_scripts": ["{}=samcli.cli.main:cli".format(cmd_name)]},
    install_requires=read_requirements("base.txt"),
    extras_require={"pre-dev": read_requirements("pre-dev.txt"), "dev": read_requirements("dev.txt")},
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Utilities",
    ],
)



"""
Code for all Package-able resources
"""
import logging
import os
import shutil
from typing import Dict, Optional, Union

import jmespath
from botocore.utils import set_value_from_jmespath

from samcli.commands.package import exceptions
from samcli.lib.package.ecr_uploader import ECRUploader
from samcli.lib.package.s3_uploader import S3Uploader
from samcli.lib.package.uploaders import Destination, Uploaders
from samcli.lib.package.utils import (
    copy_to_temp_dir,
    is_ecr_url,
    is_local_file,
    is_path_value_valid,
    is_s3_protocol_url,
    is_zip_file,
    resource_not_packageable,
    upload_local_artifacts,
    upload_local_image_artifacts,
)
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.lib.utils.resources import (
    AWS_APIGATEWAY_RESTAPI,
    AWS_APIGATEWAY_V2_API,
    AWS_APPSYNC_FUNCTIONCONFIGURATION,
    AWS_APPSYNC_GRAPHQLSCHEMA,
    AWS_APPSYNC_RESOLVER,
    AWS_CLOUDFORMATION_MODULEVERSION,
    AWS_CLOUDFORMATION_RESOURCEVERSION,
    AWS_ECR_REPOSITORY,
    AWS_ELASTICBEANSTALK_APPLICATIONVERSION,
    AWS_GLUE_JOB,
    AWS_LAMBDA_FUNCTION,
    AWS_LAMBDA_LAYERVERSION,
    AWS_SERVERLESS_API,
    AWS_SERVERLESS_FUNCTION,
    AWS_SERVERLESS_HTTPAPI,
    AWS_SERVERLESS_LAYERVERSION,
    AWS_SERVERLESS_STATEMACHINE,
    AWS_SERVERLESSREPO_APPLICATION,
    AWS_STEPFUNCTIONS_STATEMACHINE,
    METADATA_WITH_LOCAL_PATHS,
    RESOURCES_WITH_IMAGE_COMPONENT,
    RESOURCES_WITH_LOCAL_PATHS,
)

LOG = logging.getLogger(__name__)


class Resource:
    RESOURCE_TYPE: Optional[str] = None
    PROPERTY_NAME: Optional[str] = None
    PACKAGE_NULL_PROPERTY = True
    # Set this property to True in base class if you want the exporter to zip
    # up the file before uploading This is useful for Lambda functions.
    FORCE_ZIP = False
    EXPORT_DESTINATION: Destination
    ARTIFACT_TYPE: Optional[str] = None

    def __init__(self, uploaders: Uploaders, code_signer):
        self.uploaders = uploaders
        self.code_signer = code_signer

    @property
    def uploader(self) -> Union[S3Uploader, ECRUploader]:
        """
        Return the uploader matching the EXPORT_DESTINATION
        """
        return self.uploaders.get(self.EXPORT_DESTINATION)

    def export(self, resource_id, resource_dict, parent_dir):
        self.do_export(resource_id, resource_dict, parent_dir)

    def do_export(self, resource_id, resource_dict, parent_dir):
        pass

    def delete(self, resource_id, resource_dict):
        pass


class ResourceZip(Resource):
    """
    Base class representing a CloudFormation resource that can be exported
    """

    RESOURCE_TYPE: Optional[str] = None
    PROPERTY_NAME: str = ""
    PACKAGE_NULL_PROPERTY = True
    # Set this property to True in base class if you want the exporter to zip
    # up the file before uploading This is useful for Lambda functions.
    FORCE_ZIP = False
    ARTIFACT_TYPE = ZIP
    EXPORT_DESTINATION = Destination.S3

    def export(self, resource_id: str, resource_dict: Optional[Dict], parent_dir: str):
        if resource_dict is None:
            return

        if resource_not_packageable(resource_dict):
            return

        property_value = jmespath.search(self.PROPERTY_NAME, resource_dict)

        if not property_value and not self.PACKAGE_NULL_PROPERTY:
            return

        if isinstance(property_value, dict):
            LOG.debug("Property %s of %s resource is not a URL", self.PROPERTY_NAME, resource_id)
            return

        # If property is a file but not a zip file, place file in temp
        # folder and send the temp folder to be zipped
        temp_dir = None
        if is_local_file(property_value) and not is_zip_file(property_value) and self.FORCE_ZIP:
            temp_dir = copy_to_temp_dir(property_value)
            set_value_from_jmespath(resource_dict, self.PROPERTY_NAME, temp_dir)

        try:
            self.do_export(resource_id, resource_dict, parent_dir)

        except Exception as ex:
            LOG.debug("Unable to export", exc_info=ex)
            raise exceptions.ExportFailedError(
                resource_id=resource_id, property_name=self.PROPERTY_NAME, property_value=property_value, ex=ex
            )
        finally:
            if temp_dir:
                shutil.rmtree(temp_dir)

    def do_export(self, resource_id, resource_dict, parent_dir):
        """
        Default export action is to upload artifacts and set the property to
        S3 URL of the uploaded object
        If code signing configuration is provided for function/layer, uploaded artifact
        will be replaced by signed artifact location
        """
        # code signer only accepts files which has '.zip' extension in it
        # so package artifact with '.zip' if it is required to be signed
        should_sign_package = self.code_signer.should_sign_package(resource_id)
        artifact_extension = "zip" if should_sign_package else None
        uploaded_url = upload_local_artifacts(
            self.RESOURCE_TYPE,
            resource_id,
            resource_dict,
            self.PROPERTY_NAME,
            parent_dir,
            self.uploader,
            artifact_extension,
        )
        if should_sign_package:
            uploaded_url = self.code_signer.sign_package(
                resource_id, uploaded_url, self.uploader.get_version_of_artifact(uploaded_url)
            )
        set_value_from_jmespath(resource_dict, self.PROPERTY_NAME, uploaded_url)

    def delete(self, resource_id, resource_dict):
        """
        Delete the S3 artifact using S3 url referenced by PROPERTY_NAME
        """
        if resource_dict is None:
            return

        s3_info = self.get_property_value(resource_dict)
        if s3_info["Key"]:
            self.uploader.delete_artifact(s3_info["Key"], True)

    def get_property_value(self, resource_dict):
        """
        Get the s3 property value for this resource
        """
        if resource_dict is None:
            return {"Bucket": None, "Key": None}

        resource_path = jmespath.search(self.PROPERTY_NAME, resource_dict)
        # In the case where resource_path is pointing to an intrinsinc
        # ref function, sam delete will delete the stack but skip the deletion of this
        # artifact, as deletion of intrinsic ref function artifacts is not supported yet.
        # TODO: Allow deletion of S3 artifacts with intrinsic ref functions.
        if resource_path and isinstance(resource_path, str):
            return self.uploader.parse_s3_url(resource_path)
        return {"Bucket": None, "Key": None}


class ResourceImageDict(Resource):
    """
    Base class representing a CFN Image based resource that can be exported.
    """

    RESOURCE_TYPE: Optional[str] = None
    PROPERTY_NAME: Optional[str] = None
    FORCE_ZIP = False
    ARTIFACT_TYPE = IMAGE
    EXPORT_DESTINATION = Destination.ECR
    EXPORT_PROPERTY_CODE_KEY = "ImageUri"

    def export(self, resource_id, resource_dict, parent_dir):
        if resource_dict is None:
            return

        property_value = jmespath.search(self.PROPERTY_NAME, resource_dict)

        if isinstance(property_value, dict):
            LOG.debug("Property %s of %s resource is not a URL or a local image", self.PROPERTY_NAME, resource_id)
            return

        try:
            self.do_export(resource_id, resource_dict, parent_dir)

        except Exception as ex:
            LOG.debug("Unable to export", exc_info=ex)
            raise exceptions.ExportFailedError(
                resource_id=resource_id, property_name=self.PROPERTY_NAME, property_value=property_value, ex=ex
            )

    def do_export(self, resource_id, resource_dict, parent_dir):
        """
        Default export action is to upload artifacts and set the property to
        dictionary where the key is EXPORT_PROPERTY_CODE_KEY and value is set to an
        uploaded URL.
        """
        uploaded_url = upload_local_image_artifacts(
            resource_id, resource_dict, self.PROPERTY_NAME, parent_dir, self.uploader
        )
        set_value_from_jmespath(resource_dict, self.PROPERTY_NAME, {self.EXPORT_PROPERTY_CODE_KEY: uploaded_url})

    def delete(self, resource_id, resource_dict):
        """
        Delete the ECR artifact using ECR url in PROPERTY_NAME referenced by EXPORT_PROPERTY_CODE_KEY
        """
        if resource_dict is None:
            return

        remote_path = resource_dict.get(self.PROPERTY_NAME, {}).get(self.EXPORT_PROPERTY_CODE_KEY)
        # In the case where remote_path is pointing to an intrinsinc
        # ref function, sam delete will delete the stack but skip the deletion of this
        # artifact, as deletion of intrinsic ref function artifacts is not supported yet.
        # TODO: Allow deletion of ECR artifacts with intrinsic ref functions.
        if isinstance(remote_path, str) and is_ecr_url(remote_path):
            self.uploader.delete_artifact(
                image_uri=remote_path, resource_id=resource_id, property_name=self.PROPERTY_NAME
            )


class ResourceImage(Resource):
    """
    Base class representing a SAM Image based resource that can be exported.
    """

    RESOURCE_TYPE: Optional[str] = None
    PROPERTY_NAME: Optional[str] = None
    FORCE_ZIP = False
    ARTIFACT_TYPE: Optional[str] = IMAGE
    EXPORT_DESTINATION = Destination.ECR

    def export(self, resource_id, resource_dict, parent_dir):
        if resource_dict is None:
            return

        property_value = jmespath.search(self.PROPERTY_NAME, resource_dict)

        if isinstance(property_value, dict):
            LOG.debug("Property %s of %s resource is not a URL or a local image", self.PROPERTY_NAME, resource_id)
            return

        try:
            self.do_export(resource_id, resource_dict, parent_dir)

        except Exception as ex:
            LOG.debug("Unable to export", exc_info=ex)
            raise exceptions.ExportFailedError(
                resource_id=resource_id, property_name=self.PROPERTY_NAME, property_value=property_value, ex=ex
            )

    def do_export(self, resource_id, resource_dict, parent_dir):
        """
        Default export action is to upload artifacts and set the property to
        URL of the uploaded object
        """
        uploaded_url = upload_local_image_artifacts(
            resource_id, resource_dict, self.PROPERTY_NAME, parent_dir, self.uploader
        )
        set_value_from_jmespath(resource_dict, self.PROPERTY_NAME, uploaded_url)

    def delete(self, resource_id, resource_dict):
        """
        Delete the ECR artifact using ECR url referenced by property_name
        """
        if resource_dict is None:
            return

        remote_path = resource_dict.get(self.PROPERTY_NAME)
        # In the case where remote_path is pointing to an intrinsinc
        # ref function, sam delete will delete the stack but skip the deletion of this
        # artifact, as deletion of intrinsic ref function artifacts is not supported yet.
        # TODO: Allow deletion of ECR artifacts with intrinsic ref functions.
        if isinstance(remote_path, str) and is_ecr_url(remote_path):
            self.uploader.delete_artifact(
                image_uri=remote_path, resource_id=resource_id, property_name=self.PROPERTY_NAME
            )


class ResourceWithS3UrlDict(ResourceZip):
    """
    Represents CloudFormation resources that need the S3 URL to be specified as
    an dict like {Bucket: "", Key: "", Version: ""}
    """

    BUCKET_NAME_PROPERTY: Optional[str] = None
    OBJECT_KEY_PROPERTY: Optional[str] = None
    VERSION_PROPERTY: Optional[str] = None
    ARTIFACT_TYPE = ZIP
    EXPORT_DESTINATION = Destination.S3

    def do_export(self, resource_id, resource_dict, parent_dir):
        """
        Upload to S3 and set property to an dict representing the S3 url
        of the uploaded object
        """

        artifact_s3_url = upload_local_artifacts(
            self.RESOURCE_TYPE, resource_id, resource_dict, self.PROPERTY_NAME, parent_dir, self.uploader
        )

        parsed_url = S3Uploader.parse_s3_url(
            artifact_s3_url,
            bucket_name_property=self.BUCKET_NAME_PROPERTY,
            object_key_property=self.OBJECT_KEY_PROPERTY,
            version_property=self.VERSION_PROPERTY,
        )
        set_value_from_jmespath(resource_dict, self.PROPERTY_NAME, parsed_url)

    def delete(self, resource_id, resource_dict):
        """
        Delete the S3 artifact using S3 url in the dict PROPERTY_NAME
        using the bucket at BUCKET_NAME_PROPERTY and key at OBJECT_KEY_PROPERTY
        """
        if resource_dict is None:
            return

        s3_info = self.get_property_value(resource_dict)
        if s3_info["Key"]:
            self.uploader.delete_artifact(remote_path=s3_info["Key"], is_key=True)

    def get_property_value(self, resource_dict):
        """
        Get the s3 property value for this resource
        """
        if resource_dict is None:
            return {"Bucket": None, "Key": None}

        resource_path = resource_dict.get(self.PROPERTY_NAME, {})
        s3_bucket = resource_path.get(self.BUCKET_NAME_PROPERTY, None)

        key = resource_path.get(self.OBJECT_KEY_PROPERTY, None)
        # In the case where resource_path is pointing to an intrinsinc
        # ref function, sam delete will delete the stack but skip the deletion of this
        # artifact, as deletion of intrinsic ref function artifacts is not supported yet.
        # TODO: Allow deletion of S3 artifacts with intrinsic ref functions.
        if isinstance(s3_bucket, str) and isinstance(key, str):
            return {"Bucket": s3_bucket, "Key": key}
        return {"Bucket": None, "Key": None}


class ServerlessFunctionResource(ResourceZip):
    RESOURCE_TYPE = AWS_SERVERLESS_FUNCTION
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    FORCE_ZIP = True


class ServerlessFunctionImageResource(ResourceImage):
    RESOURCE_TYPE = AWS_SERVERLESS_FUNCTION
    PROPERTY_NAME = RESOURCES_WITH_IMAGE_COMPONENT[RESOURCE_TYPE][0]
    FORCE_ZIP = False


class ServerlessApiResource(ResourceZip):
    RESOURCE_TYPE = AWS_SERVERLESS_API
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    # Don't package the directory if DefinitionUri is omitted.
    # Necessary to support DefinitionBody
    PACKAGE_NULL_PROPERTY = False


class ServerlessHttpApiResource(ResourceZip):
    RESOURCE_TYPE = AWS_SERVERLESS_HTTPAPI
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    # Don't package the directory if DefinitionUri is omitted.
    # Necessary to support DefinitionBody
    PACKAGE_NULL_PROPERTY = False


class ServerlessStateMachineResource(ResourceWithS3UrlDict):
    RESOURCE_TYPE = AWS_SERVERLESS_STATEMACHINE
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    PACKAGE_NULL_PROPERTY = False
    BUCKET_NAME_PROPERTY = "Bucket"
    OBJECT_KEY_PROPERTY = "Key"
    VERSION_PROPERTY = "Version"


class GraphQLSchemaResource(ResourceZip):
    RESOURCE_TYPE = AWS_APPSYNC_GRAPHQLSCHEMA
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    # Don't package the directory if DefinitionS3Location is omitted.
    # Necessary to support Definition
    PACKAGE_NULL_PROPERTY = False


class AppSyncResolverRequestTemplateResource(ResourceZip):
    RESOURCE_TYPE = AWS_APPSYNC_RESOLVER
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    # Don't package the directory if RequestMappingTemplateS3Location is omitted.
    # Necessary to support RequestMappingTemplate
    PACKAGE_NULL_PROPERTY = False


class AppSyncResolverResponseTemplateResource(ResourceZip):
    RESOURCE_TYPE = AWS_APPSYNC_RESOLVER
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][1]
    # Don't package the directory if ResponseMappingTemplateS3Location is omitted.
    # Necessary to support ResponseMappingTemplate
    PACKAGE_NULL_PROPERTY = False


class AppSyncResolverCodeResource(ResourceZip):
    RESOURCE_TYPE = AWS_APPSYNC_RESOLVER
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][2]
    # Don't package the directory if CodeS3Location is omitted.
    # Necessary to support CodeS3Location
    PACKAGE_NULL_PROPERTY = False


class AppSyncFunctionConfigurationRequestTemplateResource(ResourceZip):
    RESOURCE_TYPE = AWS_APPSYNC_FUNCTIONCONFIGURATION
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    # Don't package the directory if RequestMappingTemplateS3Location is omitted.
    # Necessary to support RequestMappingTemplate
    PACKAGE_NULL_PROPERTY = False


class AppSyncFunctionConfigurationResponseTemplateResource(ResourceZip):
    RESOURCE_TYPE = AWS_APPSYNC_FUNCTIONCONFIGURATION
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][1]
    # Don't package the directory if ResponseMappingTemplateS3Location is omitted.
    # Necessary to support ResponseMappingTemplate
    PACKAGE_NULL_PROPERTY = False


class AppSyncFunctionConfigurationCodeResource(ResourceZip):
    RESOURCE_TYPE = AWS_APPSYNC_FUNCTIONCONFIGURATION
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][2]
    # Don't package the directory if CodeS3Location is omitted.
    # Necessary to support CodeS3Location
    PACKAGE_NULL_PROPERTY = False


class LambdaFunctionResource(ResourceWithS3UrlDict):
    RESOURCE_TYPE = AWS_LAMBDA_FUNCTION
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    BUCKET_NAME_PROPERTY = "S3Bucket"
    OBJECT_KEY_PROPERTY = "S3Key"
    VERSION_PROPERTY = "S3ObjectVersion"
    FORCE_ZIP = True


class LambdaFunctionImageResource(ResourceImage):
    RESOURCE_TYPE = AWS_LAMBDA_FUNCTION
    PROPERTY_NAME = RESOURCES_WITH_IMAGE_COMPONENT[RESOURCE_TYPE][0]
    FORCE_ZIP = True


class StepFunctionsStateMachineResource(ResourceWithS3UrlDict):
    RESOURCE_TYPE = AWS_STEPFUNCTIONS_STATEMACHINE
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    PACKAGE_NULL_PROPERTY = False
    BUCKET_NAME_PROPERTY = "Bucket"
    OBJECT_KEY_PROPERTY = "Key"
    VERSION_PROPERTY = "Version"


class ApiGatewayV2Resource(ResourceWithS3UrlDict):
    RESOURCE_TYPE = AWS_APIGATEWAY_V2_API
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    PACKAGE_NULL_PROPERTY = False
    BUCKET_NAME_PROPERTY = "Bucket"
    OBJECT_KEY_PROPERTY = "Key"
    VERSION_PROPERTY = "Version"


class ApiGatewayRestApiResource(ResourceWithS3UrlDict):
    RESOURCE_TYPE = AWS_APIGATEWAY_RESTAPI
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    PACKAGE_NULL_PROPERTY = False
    BUCKET_NAME_PROPERTY = "Bucket"
    OBJECT_KEY_PROPERTY = "Key"
    VERSION_PROPERTY = "Version"


class ElasticBeanstalkApplicationVersion(ResourceWithS3UrlDict):
    RESOURCE_TYPE = AWS_ELASTICBEANSTALK_APPLICATIONVERSION
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    BUCKET_NAME_PROPERTY = "S3Bucket"
    OBJECT_KEY_PROPERTY = "S3Key"
    VERSION_PROPERTY = None


class LambdaLayerVersionResource(ResourceWithS3UrlDict):
    RESOURCE_TYPE = AWS_LAMBDA_LAYERVERSION
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    BUCKET_NAME_PROPERTY = "S3Bucket"
    OBJECT_KEY_PROPERTY = "S3Key"
    VERSION_PROPERTY = "S3ObjectVersion"
    FORCE_ZIP = True


class ServerlessLayerVersionResource(ResourceZip):
    RESOURCE_TYPE = AWS_SERVERLESS_LAYERVERSION
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    FORCE_ZIP = True


class ServerlessRepoApplicationLicense(ResourceZip):
    RESOURCE_TYPE = AWS_SERVERLESSREPO_APPLICATION
    PROPERTY_NAME = METADATA_WITH_LOCAL_PATHS[RESOURCE_TYPE][0]
    PACKAGE_NULL_PROPERTY = False


class ServerlessRepoApplicationReadme(ResourceZip):
    RESOURCE_TYPE = AWS_SERVERLESSREPO_APPLICATION
    PROPERTY_NAME = METADATA_WITH_LOCAL_PATHS[RESOURCE_TYPE][1]
    PACKAGE_NULL_PROPERTY = False


class GlueJobCommandScriptLocationResource(ResourceZip):
    """
    Represents Glue::Job resource.
    """

    RESOURCE_TYPE = AWS_GLUE_JOB
    # Note the PROPERTY_NAME includes a '.' implying it's nested.
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[AWS_GLUE_JOB][0]


class CloudFormationModuleVersionModulePackage(ResourceZip):
    RESOURCE_TYPE = AWS_CLOUDFORMATION_MODULEVERSION
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[AWS_CLOUDFORMATION_MODULEVERSION][0]


class CloudFormationResourceVersionSchemaHandlerPackage(ResourceZip):
    RESOURCE_TYPE = AWS_CLOUDFORMATION_RESOURCEVERSION
    PROPERTY_NAME = RESOURCES_WITH_LOCAL_PATHS[AWS_CLOUDFORMATION_RESOURCEVERSION][0]


class ECRResource(Resource):
    """
    Represents CloudFormation resources ECR for deleting the ECR
    repository with the property name RepositoryName. This class is used
    only for deleting the repository and not exporting anything.
    """

    RESOURCE_TYPE = AWS_ECR_REPOSITORY
    PROPERTY_NAME = RESOURCES_WITH_IMAGE_COMPONENT[RESOURCE_TYPE][0]
    ARTIFACT_TYPE = ZIP
    EXPORT_DESTINATION = Destination.ECR

    def delete(self, resource_id, resource_dict):
        if resource_dict is None:
            return

        repository_name = self.get_property_value(resource_dict)
        # TODO: Allow deletion of ECR Repositories with intrinsic ref functions.
        if repository_name and isinstance(repository_name, str):
            self.uploader.delete_ecr_repository(physical_id=repository_name)

    def get_property_value(self, resource_dict):
        if resource_dict is None:
            return None

        return jmespath.search(self.PROPERTY_NAME, resource_dict)


RESOURCES_EXPORT_LIST = [
    ServerlessFunctionResource,
    ServerlessFunctionImageResource,
    ServerlessApiResource,
    ServerlessHttpApiResource,
    ServerlessStateMachineResource,
    GraphQLSchemaResource,
    AppSyncResolverRequestTemplateResource,
    AppSyncResolverResponseTemplateResource,
    AppSyncResolverCodeResource,
    AppSyncFunctionConfigurationRequestTemplateResource,
    AppSyncFunctionConfigurationResponseTemplateResource,
    AppSyncFunctionConfigurationCodeResource,
    ApiGatewayRestApiResource,
    ApiGatewayV2Resource,
    StepFunctionsStateMachineResource,
    LambdaFunctionResource,
    LambdaFunctionImageResource,
    ElasticBeanstalkApplicationVersion,
    ServerlessLayerVersionResource,
    LambdaLayerVersionResource,
    GlueJobCommandScriptLocationResource,
    CloudFormationModuleVersionModulePackage,
    CloudFormationResourceVersionSchemaHandlerPackage,
    ECRResource,
]

METADATA_EXPORT_LIST = [ServerlessRepoApplicationReadme, ServerlessRepoApplicationLicense]


def include_transform_export_handler(template_dict, uploader, parent_dir):
    if template_dict.get("Name", None) != "AWS::Include":
        return template_dict

    include_location = template_dict.get("Parameters", {}).get("Location", None)
    if not include_location or not is_path_value_valid(include_location) or is_s3_protocol_url(include_location):
        # `include_location` is either empty, or not a string, or an S3 URI
        return template_dict

    # We are confident at this point that `include_location` is a string containing the local path
    abs_include_location = os.path.join(parent_dir, include_location)
    if is_local_file(abs_include_location):
        template_dict["Parameters"]["Location"] = uploader.upload_with_dedup(abs_include_location)
    else:
        raise exceptions.InvalidLocalPathError(
            resource_id="AWS::Include", property_name="Location", local_path=abs_include_location
        )

    return template_dict


GLOBAL_EXPORT_DICT = {"Fn::Transform": include_transform_export_handler}



"""
The producer for the 'sam list resources' command
"""
import dataclasses
import logging
from typing import Any, Dict

from botocore.exceptions import BotoCoreError, ClientError, NoCredentialsError
from samtranslator.translator.arn_generator import NoRegionFound
from samtranslator.translator.managed_policy_translator import ManagedPolicyLoader

from samcli.commands._utils.template import get_template_data
from samcli.commands.exceptions import UserException
from samcli.commands.list.exceptions import (
    SamListLocalResourcesNotFoundError,
    SamListUnknownBotoCoreError,
    SamListUnknownClientError,
    StackDoesNotExistInRegionError,
)
from samcli.commands.local.cli_common.user_exceptions import InvalidSamTemplateException
from samcli.commands.validate.lib.exceptions import InvalidSamDocumentException
from samcli.lib.list.list_interfaces import Producer
from samcli.lib.list.resources.resources_def import ResourcesDef
from samcli.lib.providers.sam_stack_provider import SamLocalStackProvider
from samcli.lib.translate.sam_template_validator import SamTemplateValidator
from samcli.lib.utils.boto_utils import get_client_error_code
from samcli.yamlhelper import yaml_parse

LOG = logging.getLogger(__name__)

ROOT_STACK = 0


class ResourceMappingProducer(Producer):
    def __init__(
        self,
        stack_name,
        region,
        profile,
        template_file,
        cloudformation_client,
        iam_client,
        mapper,
        consumer,
    ):
        self.stack_name = stack_name
        self.region = region
        self.profile = profile
        self.template_file = template_file
        self.cloudformation_client = cloudformation_client
        self.iam_client = iam_client
        self.mapper = mapper
        self.consumer = consumer

    def get_resources_info(self):
        """
        Returns the stack resources information for the stack and raises exceptions accordingly

        Returns
        -------
            A dictionary containing information about the stack's resources
        """

        try:
            response = self.cloudformation_client.describe_stack_resources(StackName=self.stack_name)
            if "StackResources" not in response:
                return {"StackResources": []}
            return response
        except ClientError as e:
            if get_client_error_code(e) == "ValidationError":
                LOG.debug("Stack with id %s does not exist", self.stack_name)
                raise StackDoesNotExistInRegionError(stack_name=self.stack_name, region=self.region) from e
            LOG.error("ClientError Exception : %s", str(e))
            raise SamListUnknownClientError(msg=str(e)) from e
        except BotoCoreError as e:
            LOG.error("Botocore Exception : %s", str(e))
            raise SamListUnknownBotoCoreError(msg=str(e)) from e

    def get_translated_dict(self, template_file_dict: Dict[Any, Any]) -> Dict[Any, Any]:
        """
        Performs a sam translate on a template and returns the translated template in the form of a dictionary or
        raises exceptions accordingly

        Parameters
        ----------
        template_file_dict: Dict[Any, Any]
            The template in dictionary format to be translated

        Returns
        -------
        response: Dict[Any, Any]
            The dictionary representing the translated template
        """
        try:
            # Note to check if IAM can be mocked to get around doing a translate without it
            validator = SamTemplateValidator(
                template_file_dict, ManagedPolicyLoader(self.iam_client), profile=self.profile, region=self.region
            )
            translated_dict = yaml_parse(validator.get_translated_template_if_valid())
            return translated_dict
        except InvalidSamDocumentException as e:
            raise InvalidSamTemplateException(str(e)) from e
        except NoRegionFound as no_region_found_e:
            raise UserException(
                "AWS Region was not found. Please configure your region through a profile or --region option",
                wrapped_from=no_region_found_e.__class__.__name__,
            ) from no_region_found_e
        except NoCredentialsError as e:
            raise UserException(
                "AWS Credentials are required. Please configure your credentials.", wrapped_from=e.__class__.__name__
            ) from e
        except ClientError as e:
            LOG.error("ClientError Exception : %s", str(e))
            raise SamListUnknownClientError(msg=str(e)) from e

    def produce(self):
        """
        Produces the resource data to be printed
        """
        sam_template = get_template_data(self.template_file)

        translated_dict = self.get_translated_dict(template_file_dict=sam_template)

        stacks, _ = SamLocalStackProvider.get_stacks(template_file="", template_dictionary=translated_dict)
        if not stacks or not stacks[ROOT_STACK].resources:
            raise SamListLocalResourcesNotFoundError(msg="No local resources found.")
        seen_resources = set()
        resources_list = []
        if self.stack_name:
            response = self.get_resources_info()
            for deployed_resource in response["StackResources"]:
                resource_data = ResourcesDef(
                    LogicalResourceId=deployed_resource.get("LogicalResourceId"),
                    PhysicalResourceId=deployed_resource.get("PhysicalResourceId"),
                )
                resources_list.append(dataclasses.asdict(resource_data))
                seen_resources.add(deployed_resource.get("LogicalResourceId"))
            for local_resource in stacks[ROOT_STACK].resources:
                if local_resource not in seen_resources:
                    resource_data = ResourcesDef(LogicalResourceId=local_resource, PhysicalResourceId="-")
                    resources_list.append(dataclasses.asdict(resource_data))
        else:
            for local_resource in stacks[ROOT_STACK].resources:
                # Set the PhysicalID to "-" if there is no corresponding PhysicalID
                resource_data = ResourcesDef(LogicalResourceId=local_resource, PhysicalResourceId="-")
                resources_list.append(dataclasses.asdict(resource_data))
        mapped_output = self.mapper.map(resources_list)
        self.consumer.consume(mapped_output)



"""
A list of helper functions that cleanup the processing in IntrinsicResolver and IntrinsicSymbolTable
"""
from samcli.lib.intrinsic_resolver.invalid_intrinsic_exception import InvalidIntrinsicException


def verify_intrinsic_type_bool(argument, property_type="", message="", position_in_list=""):
    verify_intrinsic_type(argument, property_type, message, position_in_list, primitive_type=bool)


def verify_intrinsic_type_list(argument, property_type="", message="", position_in_list=""):
    verify_intrinsic_type(argument, property_type, message, position_in_list, primitive_type=list)


def verify_intrinsic_type_dict(argument, property_type="", message="", position_in_list=""):
    verify_intrinsic_type(argument, property_type, message, position_in_list, primitive_type=dict)


def verify_intrinsic_type_int(argument, property_type="", message="", position_in_list=""):
    # Special case since bool is a subclass of int in python
    if isinstance(argument, bool):
        raise InvalidIntrinsicException(
            message or "The {} argument to {} must resolve to a {} type".format(position_in_list, property_type, int)
        )
    verify_intrinsic_type(argument, property_type, message, position_in_list, primitive_type=int)


def verify_intrinsic_type_str(argument, property_type="", message="", position_in_list=""):
    verify_intrinsic_type(argument, property_type, message, position_in_list, primitive_type=str)


def verify_non_null(argument, property_type="", message="", position_in_list=""):
    if argument is None:
        raise InvalidIntrinsicException(
            message
            or "The {} argument to {} is missing from the intrinsic function".format(position_in_list, property_type)
        )


def verify_intrinsic_type(argument, property_type="", message="", position_in_list="", primitive_type=str):
    verify_non_null(argument, property_type, message, position_in_list)
    if not isinstance(argument, primitive_type):
        raise InvalidIntrinsicException(
            message
            or "The {} argument to {} must resolve to a {} type".format(position_in_list, property_type, primitive_type)
        )


def verify_in_bounds(objects, index, property_type=""):
    if index < 0 or index >= len(objects):
        raise InvalidIntrinsicException(
            "The index of {} resolved properties must be within the range".format(property_type)
        )


def verify_number_arguments(arguments, property_type="", num=0):
    if not len(arguments) == num:
        raise InvalidIntrinsicException(
            "The arguments to {} must have {} arguments instead of {} arguments".format(
                property_type, num, len(arguments)
            )
        )


def verify_all_list_intrinsic_type(arguments, verification_func, property_type="", message="", position_in_list=""):
    for argument in arguments:
        verification_func(argument, property_type, message, position_in_list)



"""
Row Definition and Row Modifiers to be used with `click.HelpFormatter write_dl methods`.
"""

from abc import ABC
from typing import List, NamedTuple

from click import style


class RowModifier(ABC):
    def apply(self, row, justification_length):
        raise NotImplementedError


class RowDefinition(NamedTuple):
    name: str = ""
    text: str = ""
    extra_row_modifiers: List[RowModifier] = []


class BaseLineRowModifier(RowModifier):
    def __init__(self, justification_length: int = 0):
        self.justification_length = justification_length

    def apply(self, row: RowDefinition, justification_length: int):
        return RowDefinition(
            name=row.name.ljust(justification_length if justification_length else self.justification_length),
            text=row.text.strip(),
        )


class HighlightNewRowNameModifier(RowModifier):
    NEW_TEXT = "NEW!"
    NEW_COLOR = "bright_yellow"

    def apply(self, row: RowDefinition, justification_length: int):
        return RowDefinition(
            name=style(
                f"{row.name.strip() + ' ' + self.NEW_TEXT}".ljust(justification_length),
                bold=True,
                fg=self.NEW_COLOR,
            ),
            text=row.text.strip(),
        )


class ShowcaseRowModifier(RowModifier):
    COLOR = "green"

    def apply(self, row: RowDefinition, justification_length: int):
        return RowDefinition(
            name=style(
                f"{row.name}".ljust(justification_length),
                fg=self.COLOR,
            ),
            text=style(row.text.strip(), fg=self.COLOR),
        )



"""
Terraform prepare hook implementation

This module contains the main prepare method
"""
import json
import logging
import os
from pathlib import Path
from subprocess import CalledProcessError, run
from typing import Any, Dict

from samcli.hook_packages.terraform.hooks.prepare.constants import CFN_CODE_PROPERTIES
from samcli.hook_packages.terraform.hooks.prepare.translate import translate_to_cfn
from samcli.lib.hook.exceptions import PrepareHookException
from samcli.lib.utils import osutils
from samcli.lib.utils.subprocess_utils import LoadingPatternError, invoke_subprocess_with_loading_pattern

LOG = logging.getLogger(__name__)

TERRAFORM_METADATA_FILE = "template.json"
HOOK_METADATA_KEY = "AWS::SAM::Hook"
TERRAFORM_HOOK_METADATA = {
    "HookName": "terraform",
}


def prepare(params: dict) -> dict:
    """
    Prepares a terraform application for use with the SAM CLI

    Parameters
    ----------
    params: dict
        Parameters of the IaC application

    Returns
    -------
    dict
        information of the generated metadata files
    """
    output_dir_path = params.get("OutputDirPath")

    terraform_application_dir = params.get("IACProjectPath", os.getcwd())
    if not output_dir_path:
        raise PrepareHookException("OutputDirPath was not supplied")

    LOG.debug("Normalize the project root directory path %s", terraform_application_dir)
    if not os.path.isabs(terraform_application_dir):
        terraform_application_dir = os.path.normpath(os.path.join(os.getcwd(), terraform_application_dir))
        LOG.debug("The normalized project root directory path %s", terraform_application_dir)

    LOG.debug("Normalize the OutputDirPath %s", output_dir_path)
    if not os.path.isabs(output_dir_path):
        output_dir_path = os.path.normpath(os.path.join(terraform_application_dir, output_dir_path))
        LOG.debug("The normalized OutputDirPath value is %s", output_dir_path)

    skip_prepare_infra = params.get("SkipPrepareInfra")
    metadata_file_path = os.path.join(output_dir_path, TERRAFORM_METADATA_FILE)

    if skip_prepare_infra and os.path.exists(metadata_file_path):
        LOG.info("Skipping preparation stage, the metadata file already exists at %s", metadata_file_path)
    else:
        log_msg = (
            (
                "The option to skip infrastructure preparation was provided, but AWS SAM CLI could not find "
                f"the metadata file. Preparing anyways.{os.linesep}Initializing Terraform application"
            )
            if skip_prepare_infra
            else "Initializing Terraform application"
        )
        try:
            # initialize terraform application
            LOG.info(log_msg)
            invoke_subprocess_with_loading_pattern(
                command_args={
                    "args": ["terraform", "init", "-input=false"],
                    "cwd": terraform_application_dir,
                }
            )

            # get json output of terraform plan
            LOG.info("Creating terraform plan and getting JSON output")
            with osutils.tempfile_platform_independent() as temp_file:
                invoke_subprocess_with_loading_pattern(
                    # input false to avoid SAM CLI to stuck in case if the
                    # Terraform project expects input, and customer does not provide it.
                    command_args={
                        "args": ["terraform", "plan", "-out", temp_file.name, "-input=false"],
                        "cwd": terraform_application_dir,
                    }
                )

                result = run(
                    ["terraform", "show", "-json", temp_file.name],
                    check=True,
                    capture_output=True,
                    cwd=terraform_application_dir,
                )
            tf_json = json.loads(result.stdout)

            # convert terraform to cloudformation
            LOG.info("Generating metadata file")
            cfn_dict = translate_to_cfn(tf_json, output_dir_path, terraform_application_dir)

            if cfn_dict.get("Resources"):
                _update_resources_paths(cfn_dict.get("Resources"), terraform_application_dir)  # type: ignore

            # Add hook metadata
            if not cfn_dict.get("Metadata"):
                cfn_dict["Metadata"] = {}
            cfn_dict["Metadata"][HOOK_METADATA_KEY] = TERRAFORM_HOOK_METADATA

            # store in supplied output dir
            if not os.path.exists(output_dir_path):
                os.makedirs(output_dir_path, exist_ok=True)

            LOG.info("Finished generating metadata file. Storing in %s", metadata_file_path)
            with open(metadata_file_path, "w+") as metadata_file:
                json.dump(cfn_dict, metadata_file)
        except CalledProcessError as e:
            stderr_output = str(e.stderr)

            # stderr can take on bytes or just be a plain string depending on terminal
            if isinstance(e.stderr, bytes):
                stderr_output = e.stderr.decode("utf-8")

            # one of the subprocess.run calls resulted in non-zero exit code or some OS error
            LOG.debug(
                "Error running terraform command: \n" "cmd: %s \n" "stdout: %s \n" "stderr: %s \n",
                e.cmd,
                e.stdout,
                stderr_output,
            )

            raise PrepareHookException(
                f"There was an error while preparing the Terraform application.\n{stderr_output}"
            ) from e
        except LoadingPatternError as e:
            raise PrepareHookException(f"Error occurred when invoking a process: {e}") from e
        except OSError as e:
            raise PrepareHookException(f"OSError: {e}") from e

    return {"iac_applications": {"MainApplication": {"metadata_file": metadata_file_path}}}


def _update_resources_paths(cfn_resources: Dict[str, Any], terraform_application_dir: str) -> None:
    """
    As Sam Cli and terraform handles the relative paths differently. Sam Cli handles the relative paths to be relative
    to the template, but terraform handles them to be relative to the project root directory. This Function purpose is
    to update the CFN resources paths to be absolute paths, and change relative paths to be relative to the terraform
    application root directory.

    Parameters
    ----------
    cfn_resources: dict
        CloudFormation resources
    terraform_application_dir: str
        The terraform application root directory where all paths will be relative to it
    """
    resources_attributes_to_be_updated = {
        resource_type: [property_value] for resource_type, property_value in CFN_CODE_PROPERTIES.items()
    }
    for _, resource in cfn_resources.items():
        if resource.get("Type") in resources_attributes_to_be_updated and isinstance(resource.get("Properties"), dict):
            for attribute in resources_attributes_to_be_updated[resource["Type"]]:
                original_path = resource.get("Properties", {}).get(attribute)
                if isinstance(original_path, str) and not os.path.isabs(original_path):
                    resource["Properties"][attribute] = str(Path(terraform_application_dir).joinpath(original_path))
